# -*- coding: utf-8 -*-
"""gaps Job status manager. """
import json
import logging
import time
import shutil
import datetime as dt
from pathlib import Path
from copy import deepcopy
from warnings import warn
from itertools import chain
from collections import UserDict, abc

import pandas as pd
import numpy as np

from rex.utilities import safe_json_load
from gaps.hpc import SLURM, PBS
from gaps.config import ConfigType
from gaps.utilities import recursively_update_dict, CaseInsensitiveEnum
from gaps.exceptions import gapsKeyError, gapsTypeError
from gaps.warnings import gapsWarning


logger = logging.getLogger(__name__)
JOB_STATUS_FILE = "jobstatus_{}.json"
MONITOR_PID_FILE = "monitor_pid.json"
NAMED_STATUS_FILE = "{}_status.json"
DT_FMT = "%d-%b-%Y %H:%M:%S"


class StatusField(CaseInsensitiveEnum):
    """A collection of required status fields in a status file."""

    JOB_ID = "job_id"
    JOB_STATUS = "job_status"
    PIPELINE_INDEX = "pipeline_index"
    HARDWARE = "hardware"
    OUT_FILE = "out_file"
    TIME_SUBMITTED = "time_submitted"
    TIME_START = "time_start"
    TIME_END = "time_end"
    TOTAL_RUNTIME = "total_runtime"
    RUNTIME_SECONDS = "runtime_seconds"
    MONITOR_PID = "monitor_pid"


class HardwareOption(CaseInsensitiveEnum):
    """A collection of hardware options."""

    LOCAL = "local"
    SLURM = "slurm"
    EAGLE = "eagle"
    PEREGRINE = "peregrine"
    PBS = "pbs"

    @classmethod
    def _new_post_hook(cls, obj, value):
        """Hook for post-processing after __new__"""
        obj.is_hpc = value != "local"
        if value in {"slurm", "eagle"}:
            obj.manager = SLURM()
            obj.check_status_using_job_id = (
                obj.manager.check_status_using_job_id
            )
        elif value in {"pbs", "peregrine"}:
            obj.manager = PBS()
            obj.check_status_using_job_id = (
                obj.manager.check_status_using_job_id
            )
        else:
            obj.manager = None
            obj.check_status_using_job_id = lambda *__, **___: None
        return obj


class StatusOption(CaseInsensitiveEnum):
    """A collection of job status options."""

    NOT_SUBMITTED = "not submitted"
    SUBMITTED = "submitted"
    RUNNING = "running"
    SUCCESSFUL = "successful"
    FAILED = "failed"
    COMPLETE = "complete"

    @classmethod
    def _new_post_hook(cls, obj, value):
        """Hook for post-processing after __new__"""
        verb = "has" if value == "failed" else "is"
        obj.with_verb = " ".join([verb, value])
        obj.is_processing = value in {"submitted", "running"}
        return obj


# pylint: disable=too-few-public-methods
class HardwareStatusRetriever:
    """Query hardware for job status."""

    def __init__(self, subprocess_manager=None):
        """Initialize `HardwareStatusRetriever`.

        Parameters
        ----------
        subprocess_manager : PBS | SLURM | None, optional
            Optional initialized subprocess manager to use to check job
            statuses. This can be input with cached queue data to avoid
            constantly querying the HPC. By default, `None`.
        """
        self.subprocess_manager = subprocess_manager

    def __getitem__(self, key):
        """Get the job status using pre-defined hardware-specific methods."""
        job_id, hardware = key
        if not job_id:
            return None

        if self.subprocess_manager is not None:
            return self.subprocess_manager.check_status_using_job_id(job_id)

        return _validate_hardware(hardware).check_status_using_job_id(job_id)


class Status(UserDict):
    """Base class for data pipeline health and status information.

    This class facilitates the creating of individual job status files
    which the main `Status` object can collect.
    """

    _DF_COLUMNS = [
        StatusField.JOB_STATUS.value,
        StatusField.PIPELINE_INDEX.value,
        StatusField.JOB_ID.value,
        StatusField.TIME_SUBMITTED.value,
        StatusField.TIME_START.value,
        StatusField.TIME_END.value,
        StatusField.TOTAL_RUNTIME.value,
        StatusField.HARDWARE.value,
    ]

    def __init__(self, status_dir):
        """Initialize `Status`.

        Parameters
        ----------
        status_dir : path-like
            Directory containing zero or more job json status files.
        """
        super().__init__()
        self.dir = Path(status_dir).expanduser().resolve()
        self._validate_dir()
        self.name = self.dir.name
        self._fpath = self.dir / NAMED_STATUS_FILE.format(self.name)
        self.data = _load(self._fpath)

    def _validate_dir(self):
        """Validate that the directory name is not a config file type."""
        for file_type in ConfigType.members_as_str():
            if self.dir.name.endswith(f".{file_type}"):
                raise gapsTypeError(
                    f"Need a directory containing a status {file_type}, "
                    f"not a status {file_type}: {self.dir!r}"
                )

    @property
    def job_ids(self):
        """list: Flat list of job ids."""
        return _get_attr_flat_list(self.data, key=StatusField.JOB_ID)

    def as_df(self, commands=None, index_name="job_name", include_cols=None):
        """Format status as pandas DataFrame.

        Parameters
        ----------
        commands : container, optional
            A container of command names to collect. If `None`, all
            commands in the status file are collected.
            By default, `None`.
        index_name : str, optional
            Name to assign to index of DataFrame.
            By default, `"job_name"`.

        Returns
        -------
        pd.DataFrame
            Pandas DataFrame containing status information.
        """
        include_cols = include_cols or []
        output_cols = self._DF_COLUMNS + list(include_cols)

        self.update_from_all_job_files(purge=False)
        if not self.data:
            return pd.DataFrame(columns=output_cols)

        requested_commands = commands or self.keys()
        commands = []
        for command, status in self.items():
            if command not in requested_commands:
                continue
            try:
                command_index = status.pop(StatusField.PIPELINE_INDEX, None)
            except (AttributeError, TypeError):
                continue
            if not status:
                status = {command: {}}
            try:
                command_df = pd.DataFrame(status).T
            except ValueError:
                continue
            command_df[f"{StatusField.PIPELINE_INDEX}"] = command_index
            commands.append(command_df)
        try:
            command_df = pd.concat(commands, sort=False)
        except ValueError:
            return pd.DataFrame(columns=output_cols)

        for field in chain(StatusField, include_cols):
            if field not in command_df.columns:
                command_df[f"{field}"] = np.nan

        command_df.loc[
            command_df[StatusField.JOB_STATUS].isna(),
            StatusField.JOB_STATUS.value,
        ] = StatusOption.NOT_SUBMITTED.value

        command_df = _add_elapsed_time(command_df)

        command_df.index.name = index_name
        command_df = command_df[output_cols]
        return command_df

    def reload(self):
        """Re-load the data from disk."""
        self.data = _load(self._fpath)

    def dump(self):
        """Dump status json w/ backup file in case process gets killed."""

        self._fpath.parent.mkdir(exist_ok=True)

        backup = self._fpath.name.replace(".json", "_backup.json")
        backup = self._fpath.parent / backup
        if self._fpath.exists():
            shutil.copy(self._fpath, backup)

        with open(self._fpath, "w") as status:
            json.dump(self.data, status, indent=4, separators=(",", ": "))

        backup.unlink(missing_ok=True)

    def update_from_all_job_files(self, purge=True):
        """Update status from all single-job job status files.

        This method loads all single-job status files in the target
        directory and updates the `Status` object with the single-job
        statuses.

        Parameters
        ----------
        purge : bool, optional
            Option to purge the individual status files.
            By default,`True`.

        Returns
        -------
        `Status`
            This instance of `Status` with updated job properties.
        """
        file_pattern = JOB_STATUS_FILE.format("*")
        for file_ in Path(self.dir).glob(file_pattern):
            status = _safe_load(file_, purge=purge)
            self.data = recursively_update_dict(self.data, status)

        monitor_pid_file = Path(self.dir) / MONITOR_PID_FILE
        if monitor_pid_file.exists():
            monitor_pid_info = _safe_load(monitor_pid_file, purge=purge)
            self.data.update(monitor_pid_info)

        if purge:
            self.dump()

        return self

    def update_job_status(
        self, command, job_name, hardware_status_retriever=None
    ):
        """Update single-job job status from single-job job status file.

        If the status for a given command/job name combination is not
        found, the status object remains unchanged.

        Parameters
        ----------
        command : str
            CLI command that the job belongs to.
        job_name : str
            Unique job name identification.
        hardware_status_retriever : `HardwareStatusRetriever`, optional
            Hardware status retriever. By default, `None`, which creates
            an instance internally.
        """
        hardware_status_retriever = (
            hardware_status_retriever or HardwareStatusRetriever()
        )
        # look for job completion file.
        current = _load_job_file(self.dir, job_name)

        # Update status data dict and file if job file was found
        if current is not None:
            self.data = recursively_update_dict(self.data, current)
            self.dump()

        # check job status via hardware if job file not found.
        elif command in self.data:
            # job exists
            if job_name in self.data[command]:
                job_data = self.data[command][job_name]

                # init defaults in case job/command not in status file yet
                previous = job_data.get(StatusField.JOB_STATUS, None)
                job_id = job_data.get(StatusField.JOB_ID, None)
                job_hardware = job_data.get(StatusField.HARDWARE, None)

                # get job status from hardware
                current = hardware_status_retriever[job_id, job_hardware]

                # No current status and job was not successful: failed!
                if current is None and previous != StatusOption.SUCCESSFUL:
                    job_data[StatusField.JOB_STATUS] = StatusOption.FAILED

                # do not overwrite a successful or failed job status.
                elif (
                    current != previous
                    and previous not in StatusOption.members_as_str()
                ):
                    job_data[StatusField.JOB_STATUS] = current

            # job does not yet exist
            else:
                self.data[command][job_name] = {}

    def _retrieve_job_status(
        self, command, job_name, hardware_status_retriever
    ):
        """Update and retrieve job status."""
        if job_name.endswith(".h5"):
            job_name = job_name.replace(".h5", "")

        self.update_job_status(command, job_name, hardware_status_retriever)

        try:
            job_data = self[command][job_name]
        except KeyError:
            return None

        return job_data.get(StatusField.JOB_STATUS)

    @staticmethod
    def record_monitor_pid(status_dir, pid):
        """Make a json file recording the PID of the monitor process.

        Parameters
        ----------
        status_dir : path-like
            Directory to put json status file.
        pid : int
            PID of the monitoring process.
        """
        pid = {StatusField.MONITOR_PID: pid}
        fpath = Path(status_dir) / MONITOR_PID_FILE
        with open(fpath, "w") as pid_file:
            json.dump(
                pid,
                pid_file,
                sort_keys=True,
                indent=4,
                separators=(",", ": "),
            )

    @staticmethod
    def make_single_job_file(status_dir, command, job_name, attrs):
        """Make a json file recording the status of a single job.

        This method should primarily be used by HPC nodes to mark the
        status of individual jobs.

        Parameters
        ----------
        status_dir : path-like
            Directory to put json status file.
        command : str
            CLI command that the job belongs to.
        job_name : str
            Unique job name identification.
        attrs : dict
            Dictionary of job attributes that represent the job status
            attributes.
        """
        if job_name.endswith(".h5"):
            job_name = job_name.replace(".h5", "")

        status = {command: {job_name: attrs}}
        fpath = Path(status_dir) / JOB_STATUS_FILE.format(job_name)
        with open(fpath, "w") as job_status:
            json.dump(
                status,
                job_status,
                sort_keys=True,
                indent=4,
                separators=(",", ": "),
            )

    @classmethod
    def mark_job_as_submitted(
        cls, status_dir, command, job_name, replace=False, job_attrs=None
    ):
        """Mark a job in the status json as "submitted".

        Status json does not have to exist - it is created if missing.
        If it exists, the job will only be updated if it does not
        exist (i.e. not submitted), unless replace is set to `True`.

        Parameters
        ----------
        status_dir : path-like
            Directory containing json status file.
        command : str
            CLI command that the job belongs to.
        job_name : str
            Unique job name identification.
        replace : bool, optional
            Flag to force replacement of pre-existing job status.
            By default, `False`.
        job_attrs : dict, optional
            Job attributes. Should include 'job_id' if running on HPC.
            By default, `None`.
        """
        if job_name.endswith(".h5"):
            job_name = job_name.replace(".h5", "")

        obj = cls(status_dir)

        job_attrs = job_attrs or {}
        hardware = job_attrs.get(StatusField.HARDWARE)
        try:
            job_on_hpc = HardwareOption(hardware).is_hpc
        except ValueError:
            job_on_hpc = False

        if job_on_hpc and StatusField.JOB_ID not in job_attrs:
            msg = (
                f'Key "job_id" should be in kwargs for {job_name!r} if '
                f"adding job from an HPC node."
            )
            warn(msg, gapsWarning)

        exists = obj.job_exists(status_dir, job_name, command=command)

        if exists and not replace:
            return

        job_status = job_attrs.get(StatusField.JOB_STATUS)
        if job_status != StatusOption.SUBMITTED:
            if job_status is not None:
                msg = (
                    f"Attempting to mark a job as submitted but included a "
                    f"{StatusField.JOB_STATUS} value of {job_status!r} in the "
                    f"job_attrs dictionary! Setting the job status to "
                    f"{StatusOption.SUBMITTED!r} before writing."
                )
                warn(msg, gapsWarning)
            job_attrs[StatusField.JOB_STATUS] = StatusOption.SUBMITTED

        if command not in obj.data:
            obj.data[command] = {job_name: job_attrs}
        else:
            obj.data[command][job_name] = job_attrs

        obj.dump()

    @classmethod
    def job_exists(cls, status_dir, job_name, command=None):
        """Check whether a job exists and return a bool.

        This method will return `True` if the job name is found as a
        key in the dictionary under the `command` keyword, or any
        command if a `None` value is passed.

        Parameters
        ----------
        status_dir : str
            Directory containing json status file.
        job_name : str
            Unique job name identification.
        command : str, optional
            CLI command that the job belongs to. By default, `None`,
            which checks all commands for the job name.

        Returns
        -------
        exists : bool
            `True` if the job exists in the status json.
        """
        if job_name.endswith(".h5"):
            job_name = job_name.replace(".h5", "")

        obj = cls(status_dir).update_from_all_job_files(purge=False)
        if not obj.data:
            return False
        if command is not None:
            jobs = [obj.data.get(command)]
        else:
            jobs = obj.data.values()

        for job in jobs:
            if not job:
                continue
            for name in job.keys():
                if name == job_name:
                    return True

        return False

    @classmethod
    def retrieve_job_status(
        cls,
        status_dir,
        command,
        job_name,
        subprocess_manager=None,
    ):
        """Update and retrieve job status.

        Parameters
        ----------
        status_dir : str
            Directory containing json status file.
        command : str
            CLI command that the job belongs to.
        job_name : str
            Unique job name identification.
        subprocess_manager : None | SLURM, optional
            Optional initialized subprocess manager to use to check job
            statuses. This can be input with cached queue data to avoid
            constantly querying the HPC.

        Returns
        -------
        status : str | None
            Status string or `None` if job/command not found.
        """
        hsr = HardwareStatusRetriever(subprocess_manager)
        return cls(status_dir)._retrieve_job_status(command, job_name, hsr)

    @classmethod
    def parse_command_status(
        cls,
        status_dir,
        command,
        key=StatusField.OUT_FILE,
    ):
        """Parse key from job status(es) from the given command.

        Parameters
        ----------
        status_dir : path-like
            Directory containing the status file to parse.
        command : str
            Target CLI command to parse.
        key : StatusField | str, optional
            Parsing target of previous command. By default,
            `StatusField.OUT_FILE`.

        Returns
        -------
        list
            Arguments parsed from the status file in status_dir from
            the input command. This list is empty if the `key` is not
            found in the job status, or if the command does not exist in
            status.
        """
        status = cls(status_dir).update_from_all_job_files(purge=False)
        command_status = status.get(command, {})
        command_status.pop(StatusField.PIPELINE_INDEX, None)
        return _get_attr_flat_list(command_status, key=key)


class StatusUpdates:
    """Context manager to track run function progress.

    When this context is entered, a status file is written for the given
    command/job combination, with the given job attributes. The job
    status is set to "running", and the start time is recorded. When the
    context is exited, another status file is written with the end time
    and total runtime values added. The status is also set to
    "successful", unless an uncaught exception was raised during the
    function runtime, in which case the status is set to "failed". If
    the `out_file` attribute of this context manager is set before the
    context is exited, that value is also written to the status file.
    """

    def __init__(self, directory, command, job_name, job_attrs):
        """Initialize `StatusUpdates`.

        Parameters
        ----------
        directory : path-like
            Directory to write status files to.
        command : str
            Name of the command being run.
        job_name : str
            Name of the job being run.
        job_attrs : dict
            A dictionary containing job attributes that should be
            written to the status file.
        """
        self.directory = directory
        self.command = command
        self.job_name = job_name
        self.job_attrs = deepcopy(job_attrs)
        self.start_time = None
        self.out_file = None

    def __enter__(self):
        self.start_time = dt.datetime.now()
        self.job_attrs.update(
            {
                StatusField.JOB_STATUS: StatusOption.RUNNING,
                StatusField.TIME_START: self.start_time.strftime(DT_FMT),
            }
        )
        Status.make_single_job_file(
            self.directory,
            command=self.command,
            job_name=self.job_name,
            attrs=self.job_attrs,
        )
        return self

    def __exit__(self, exc_type, exc, traceback):
        end_time = dt.datetime.now()
        time_elapsed_s = (end_time - self.start_time).total_seconds()
        time_elapsed = _elapsed_time_as_str(time_elapsed_s)
        self.job_attrs.update(
            {
                StatusField.TIME_END: end_time.strftime(DT_FMT),
                StatusField.TOTAL_RUNTIME: time_elapsed,
                StatusField.RUNTIME_SECONDS: time_elapsed_s,
            }
        )
        if exc is None:
            self.job_attrs.update(
                {
                    StatusField.OUT_FILE: self.out_file,
                    StatusField.JOB_STATUS: StatusOption.SUCCESSFUL,
                }
            )
        else:
            self.job_attrs.update(
                {StatusField.JOB_STATUS: StatusOption.FAILED}
            )
        Status.make_single_job_file(
            self.directory,
            command=self.command,
            job_name=self.job_name,
            attrs=self.job_attrs,
        )


def _elapsed_time_as_str(seconds_elapsed):
    """Format elapsed time into human readable string"""
    days, seconds = divmod(int(seconds_elapsed), 24 * 3600)
    minutes, seconds = divmod(seconds, 60)
    hours, minutes = divmod(minutes, 60)
    time_str = f"{hours:d}:{minutes:02d}:{seconds:02d}"
    if days:
        time_str = f"{days:d} day{'s' if abs(days) != 1 else ''}, {time_str}"
    return time_str


def _add_elapsed_time(status_df):
    """Add elapsed time to status DataFrame"""
    mask = (
        ~status_df[StatusField.TIME_START].isna()
        & status_df[StatusField.TIME_END].isna()
    )
    start_times = status_df.loc[mask, StatusField.TIME_START]
    start_times = pd.to_datetime(start_times, format=DT_FMT)
    elapsed_times = dt.datetime.now() - start_times
    elapsed_times = elapsed_times.apply(lambda dt: dt.total_seconds())
    elapsed_times = elapsed_times.apply(_elapsed_time_as_str)
    elapsed_times = elapsed_times.apply(lambda time_str: f"{time_str} (r)")
    status_df.loc[mask, StatusField.TOTAL_RUNTIME] = elapsed_times
    return status_df


def _load(fpath):
    """Load status json."""
    if fpath.is_file():
        return safe_json_load(fpath.as_posix())
    return {}


def _load_job_file(status_dir, job_name, purge=True):
    """Load a single-job job status file in the target status_dir."""
    status_dir = Path(status_dir)
    status_fname = status_dir / JOB_STATUS_FILE.format(job_name)
    if status_fname not in status_dir.glob("*"):
        return None
    return _safe_load(status_fname, purge=purge)


def _safe_load(file_path, purge=True):
    """Safe load json file and purge if needed."""
    # wait one second to make sure file is finished being written
    time.sleep(0.01)
    status = safe_json_load(file_path.as_posix())
    if purge:
        file_path.unlink()
    return status


def _get_attr_flat_list(inp, key=StatusField.JOB_ID):
    """Get all job attribute values from the status data dict."""

    out = []
    if not isinstance(inp, abc.Mapping):
        return out

    if key in inp:
        out = _combine_iterables_and_numerics(out, inp[key])
    else:
        for val in inp.values():
            temp = _get_attr_flat_list(val, key=key)
            out = _combine_iterables_and_numerics(out, temp)

    return out


def _combine_iterables_and_numerics(iterable, new_vals):
    """Combine new_vals into the "iterable" list"""
    try:
        return iterable + new_vals
    except TypeError:
        pass

    return iterable + [new_vals]


def _validate_hardware(hardware):
    """Verify that the selected hardware is a valid option."""
    try:
        return HardwareOption(hardware)
    except ValueError as err:
        msg = (
            f"Requested hardware ({hardware!r}) not recognized! "
            f"Available options are: {HardwareOption.members_as_str()}."
        )
        raise gapsKeyError(msg) from err



# -*- coding: utf-8 -*-
"""
GAPs pipeline CLI entry points.
"""
import os
import sys
import logging
from pathlib import Path

import click

from gaps import Pipeline
from gaps.config import ConfigType
from gaps.cli.documentation import _pipeline_command_help
from gaps.cli.command import _WrappedCommand
from gaps.status import Status
from gaps.exceptions import gapsExecutionError


logger = logging.getLogger(__name__)


def _can_run_background():
    """Determine if GAPs can execute the pipeline in the background."""
    return hasattr(os, "setsid") and hasattr(os, "fork")


def template_pipeline_config(commands):
    """Generate a template pipeline config based on the commands."""
    _pipeline = []
    for command in commands:
        command_name_config = command.name.replace("-", "_")
        sample_config_filename = (
            f"./config_{command_name_config}.{ConfigType.JSON}"
        )
        _pipeline.append({command.name: sample_config_filename})
    return {
        "pipeline": _pipeline,
        "logging": {"log_file": None, "log_level": "INFO"},
    }


@click.pass_context
def pipeline(ctx, config_file, cancel, monitor, background=False):
    """Execute multiple steps in an analysis pipeline."""

    if config_file is None:
        config_file = [
            fp
            for fp in Path(".").glob("*")
            if fp.is_file() and "pipeline" in fp.name
        ]
        if len(config_file) != 1:
            msg = (
                f"Could not determine config file - multiple (or no) files "
                f" detected with 'pipeline' in the name exist: {config_file}"
            )
            raise gapsExecutionError(msg)

        config_file = config_file[0]

    if cancel:
        Pipeline.cancel_all(config_file)
        return
    if background:
        if not _can_run_background():
            msg = (
                "Cannot run pipeline in background on system that does not "
                "implement os.fork and os.setsid"
            )
            raise gapsExecutionError(msg)
        ctx.obj["LOG_STREAM"] = False
        _kickoff_background(config_file)

    Pipeline.run(config_file, monitor=monitor)


# pylint: disable=no-member
def _kickoff_background(config_file):  # pragma: no cover
    """Kickoff a child process that runs pipeline in the background."""
    pid = os.fork()
    if pid == 0:
        os.setsid()  # This creates a new session
        Pipeline.run(config_file, monitor=True)
    else:
        Status.record_monitor_pid(Path(config_file).parent, pid)
        click.echo(
            f"Kicking off pipeline job in the background. Monitor PID: {pid}"
        )
        sys.exit()


def pipeline_command(template_config):
    """Generate a pipeline command."""
    params = [
        click.Option(
            param_decls=["--config_file", "-c"],
            default=None,
            help=_pipeline_command_help(template_config),
        ),
        click.Option(
            param_decls=["--cancel"],
            is_flag=True,
            help="Flag to cancel all jobs associated with a given pipeline.",
        ),
        click.Option(
            param_decls=["--monitor"],
            is_flag=True,
            help="Flag to monitor pipeline jobs continuously. Default is not "
            "to monitor (kick off jobs and exit).",
        ),
    ]
    if _can_run_background():
        params += [
            click.Option(
                param_decls=["--background"],
                is_flag=True,
                help="Flag to monitor pipeline jobs continuously in the "
                "background. Note that the stdout/stderr will not be "
                "captured, but you can set a pipeline 'log_file' to "
                "capture logs.",
            ),
        ]

    return _WrappedCommand(
        "pipeline",
        context_settings=None,
        callback=pipeline,
        params=params,
        help="""Execute multiple steps in an analysis pipeline""",
        epilog=None,
        short_help=None,
        options_metavar="[OPTIONS]",
        add_help_option=True,
        no_args_is_help=False,
        hidden=False,
        deprecated=False,
    )



# -*- coding: utf-8 -*-
"""
Generation CLI entry points.
"""
from functools import partial

import click

from gaps import Pipeline
from gaps.cli.batch import batch_command
from gaps.cli.templates import template_command
from gaps.cli.pipeline import pipeline_command, template_pipeline_config
from gaps.cli.collect import collect
from gaps.cli.config import from_config
from gaps.cli.command import CLICommandConfiguration, _WrappedCommand
from gaps.cli.preprocessing import preprocess_collect_config
from gaps.cli.status import status_command


class _CLICommandGenerator:
    """Generate commands from a list of configurations."""

    def __init__(self, command_configs):
        """

        Parameters
        ----------
        command_configs : list of :class:`CLICommandConfiguration`
            List of command configs to convert to click commands.
        """
        self.command_configs = command_configs
        self.commands, self.template_configs = [], {}

    def add_collect_command_configs(self):
        """Add collect command if the function is split spatially."""
        all_commands = []
        for command_configuration in self.command_configs:
            all_commands.append(command_configuration)
            if command_configuration.is_split_spatially:
                collect_configuration = CLICommandConfiguration(
                    name=f"collect-{command_configuration.name}",
                    function=collect,
                    split_keys=[("_out_path", "_pattern")],
                    config_preprocessor=preprocess_collect_config,
                )
                all_commands.append(collect_configuration)
        self.command_configs = all_commands
        return self

    def convert_to_commands(self):
        """Convert all of the command configs into click commands."""
        for command_config in self.command_configs:
            func_doc = command_config.function_documentation
            name = command_config.name
            params = [
                click.Option(
                    param_decls=["--config_file", "-c"],
                    required=True,
                    type=click.Path(exists=True),
                    help=func_doc.config_help(name),
                )
            ]

            command = _WrappedCommand(
                name,
                context_settings=None,
                callback=partial(
                    from_config,
                    command_config=command_config,
                ),
                params=params,
                help=func_doc.command_help(name),
                epilog=None,
                short_help=None,
                options_metavar="[OPTIONS]",
                add_help_option=True,
                no_args_is_help=True,
                hidden=False,
                deprecated=False,
            )
            self.commands.append(command)
            Pipeline.COMMANDS[name] = command
            self.template_configs[name] = func_doc.template_config
        return self

    def add_pipeline_command(self):
        """Add pipeline command, which includes the previous commands."""
        tpc = template_pipeline_config(self.command_configs)
        pipeline = pipeline_command(tpc)
        self.commands = [pipeline] + self.commands
        self.template_configs["pipeline"] = tpc
        return self

    def add_batch_command(self):
        """Add the batch command."""
        self.commands.append(batch_command())
        return self

    def add_status_command(self):
        """Add the status command."""
        self.commands.append(status_command())
        return self

    def add_template_command(self):
        """Add the config template command."""
        self.commands.append(template_command(self.template_configs))
        return self

    def generate(self):
        """Generate a list of click commands from input configurations."""
        return (
            self.add_collect_command_configs()
            .convert_to_commands()
            .add_pipeline_command()
            .add_batch_command()
            .add_status_command()
            .add_template_command()
            .commands
        )


def make_cli(commands, info=None):
    """Create a pipeline CLI to split execution across HPC nodes.

    This function generates a CLI for a package based on the input
    command configurations. Each command configuration is based around
    a function that is to be executed on one or more nodes.

    Parameters
    ----------
    commands : list of :class:`~gaps.cli.command.CLICommandConfiguration`
        List of command configs to convert to click commands. See the
        :class:`~gaps.cli.command.CLICommandConfiguration` documentation
        for a description of the input options. Each command
        configuration is converted into a subcommand. Any command
        configuration with ``project_points`` in the `split_keys`
        argument will get a corresponding ``collect-{command name}``
        command that collects the outputs of the spatially-distributed
        command.
    info : dict, optional
        A dictionary that contains optional info about the calling
        program to include in the CLI. Allowed keys include:

            name : str
                Name of program to display at the top of the CLI help.
            version : str
                Include program version with a ``--version`` CLI option.

        By default, ``None``.

    Returns
    -------
    click.Group
        A group containing the requested subcommands. This can be used
        as the entrypoint to the package CLI.
    """
    info = info or {}
    command_generator = _CLICommandGenerator(commands)

    options = [
        click.Option(
            param_decls=["-v", "--verbose"],
            is_flag=True,
            help="Flag to turn on debug logging. Default is not verbose.",
        ),
    ]
    prog_name = [info["name"]] if "name" in info else []
    main = click.Group(
        help=" ".join(prog_name + ["Command Line Interface"]),
        params=options,
        callback=_main_cb,
        commands=command_generator.generate(),
    )
    version = info.get("version")
    if version is not None:
        main = click.version_option(version=version)(main)

    return main


@click.pass_context
def _main_cb(ctx, verbose):
    """Set the obj and verbose settings of the commands."""
    ctx.ensure_object(dict)
    ctx.obj["VERBOSE"] = verbose



# -*- coding: utf-8 -*-
"""
CLI documentation utilities.
"""
from copy import deepcopy
from inspect import signature

from numpydoc.docscrape import NumpyDocString

from gaps.config import config_as_str_for_docstring, ConfigType


DEFAULT_EXEC_VALUES = {
    "option": "local",
    "allocation": "[REQUIRED IF ON HPC]",
    "walltime": "[REQUIRED IF ON HPC]",
    "memory": None,
    "nodes": 1,
    "queue": None,
    "feature": None,
    "conda_env": None,
    "module": None,
    "sh_script": None,
}

CONFIG_TYPES = [
    ConfigType.JSON,
    ConfigType.YAML,
    ConfigType.TOML,
]

PIPELINE_CONFIG_DOC = """
Path to the "pipeline" configuration file. This argument can be left out,
but *one and only one file* with the name "pipeline" should exist in the
directory and contain the config information. Below is a sample template config

.. tabs::

    .. tab:: JSON
        ::

            {template_json_config}

    .. tab:: YAML
        ::

            {template_yaml_config}

    .. tab:: TOML
        ::

            {template_toml_config}


Parameters
----------
pipeline : list of dicts
    A list of dictionaries, where each dictionary represents one step
    in the pipeline. Each dictionary should have one key-value pair,
    where the key is the name of the CLI command to run, and the value
    is the path to a config file containing the configuration for that
    command.
logging : dict, optional
    Dictionary containing keyword-argument pairs to pass to
    :func:`gaps.log.init_logger`.

"""

BATCH_CONFIG_DOC = """
Path to the "batch" configuration file. Below is a sample template config

.. tabs::

    .. tab:: JSON
        ::

            {template_json_config}

    .. tab:: YAML
        ::

            {template_yaml_config}

    .. tab:: TOML
        ::

            {template_toml_config}


Parameters
----------
pipeline_config : str
    Path to the pipeline configuration defining the commands to run for
    every parametric set.
sets : list of dicts
    A list of dictionaries, where each dictionary defines a "set" of
    parametric runs. Each dictionary should have the following keys:

        args : dict
            A dictionary defining the arguments across all input
            configuration files to parameterize. Each argument to be
            parametrized should be a key in this dictionary, and the
            value should be a list of the parameter values to run for
            this argument.

            .. tabs::

                .. tab:: JSON
                    ::

                        {sample_json_args_dict}

                .. tab:: YAML
                    ::

                        {sample_yaml_args_dict}

                .. tab:: TOML
                    ::

                        {sample_toml_args_dict}


            This example would run a total of six pipelines, one with
            each of the following arg combinations:
            ::

                input_constant_1=18.20, path_to_a_file="/first/path.h5"
                input_constant_1=18.20, path_to_a_file="/second/path.h5"
                input_constant_1=18.20, path_to_a_file="/third/path.h5"
                input_constant_1=19.04, path_to_a_file="/first/path.h5"
                input_constant_1=19.04, path_to_a_file="/second/path.h5"
                input_constant_1=19.04, path_to_a_file="/third/path.h5"


            Remember that the keys in the ``args`` dictionary should be
            part of (at least) one of your other configuration files.
        files : list
            A list of paths to the configuration files that contain the
            arguments to be updated for every parametric run. Arguments
            can be spread out over multiple files. For example:

            .. tabs::

                .. tab:: JSON
                    ::

                        {sample_json_files}

                .. tab:: YAML
                    ::

                        {sample_yaml_files}

                .. tab:: TOML
                    ::

                        {sample_toml_files}


        set_tag : str, optional
            Optional string defining a set tag that will prefix each
            job tag for this set. This tag does not need to include an
            underscore, as that is provided during concatenation.


"""

CONFIG_DOC = """
Path to the ``{name}`` configuration file. Below is a sample template config

.. tabs::

    .. tab:: JSON
        ::

            {template_json_config}

    .. tab:: YAML
        ::

            {template_yaml_config}

    .. tab:: TOML
        ::

            {template_toml_config}


{docstring}

Note that you may remove any keys with a ``null`` value if you do not intend to update them yourself.
"""
COMMAND_DOC = """
Execute the ``{name}`` step from a config file.

{desc}

"""
EXEC_CONTROL_DOC = """
Parameters
----------
    execution_control : dict
        Dictionary containing execution control arguments. Allowed
        arguments are:

        :option: ({{'local', 'slurm', 'eagle', 'pbs', 'peregrine'}})
                 Hardware run option. 'eagle' and 'peregrine' are aliases for
                 'slurm and 'pbs', respectively.
        :allocation: (str) HPC project (allocation) handle.
        :walltime: (int) Node walltime request in hours.
        :memory: (int, optional) Node memory request in GB. Default is not to
                 specify.{n}{mw}
        :queue: (str, optional; PBS ONLY) HPC queue to submit job to.
                Examples include: 'debug', 'short', 'batch', 'batch-h',
                'long', etc. By default, `None`, which uses `test_queue`.
        :feature: (str, optional) Additional flags for SLURM job
                  (e.g. "--qos=high", "-p debug", etc). Default is not
                  to specify.
        :conda_env: (str, optional) Name of conda environment to activate.
                    Default is not to load any environments.
        :module: (str, optional) Module to load. Default is not to load any
                 modules.
        :sh_script: (str, optional) Extra shell script to run before
                    command call. Default is not to run any scripts.

        Only the "option" input is required for local execution. For
        execution on the HPC, the allocation and walltime are also
        required. All other options are populated with default values,
        as seen above.

"""
NODES_DOC = (
    "\n        :nodes: (int, optional) Number of nodes to split the project "
    "\n                points across. Note that the total number of requested "
    "\n                nodes for a job may be larger than this value if the "
    "\n                command splits across other inputs (e.g. analysis "
    "\n                years) Default is 1."
)
MW_DOC = "\n        :max_workers: ({type}) {desc}"


class FunctionDocumentation:
    """Generate documentation for a function."""

    REQUIRED_TAG = "[REQUIRED]"

    def __init__(self, *functions, skip_params=None, is_split_spatially=False):
        """
        Parameters
        ----------
        *functions : callables
            Functions that comprise a single command for which to
            generate documentation. **IMPORTANT** The extended summary
            will be pulled form the first function only!
        skip_params : set, optional
            Set of parameter names (str) to exclude from documentation.
            Typically this is because the user would not explicitly have
            to specify these. By default, `None`.
        is_split_spatially : bool, optional
            Flag indicating wether or not this function is split
            spatially across nodes. If `True`, a "nodes" option is added
            to the execution control block of the generated
            documentation. By default, `False`.
        """
        self.signatures = [signature(func) for func in functions]
        self.docs = [NumpyDocString(func.__doc__ or "") for func in functions]
        self.param_docs = {
            p.name: p for doc in self.docs for p in doc["Parameters"]
        }
        self.skip_params = set() if skip_params is None else set(skip_params)
        self.skip_params |= {"cls", "self", "max_workers"}
        self.is_split_spatially = is_split_spatially

    @property
    def default_exec_values(self):
        """dict: Default "execution_control" config."""
        exec_vals = deepcopy(DEFAULT_EXEC_VALUES)
        if not self.is_split_spatially:
            exec_vals.pop("nodes", None)
        if self.max_workers_in_func_signature:
            exec_vals["max_workers"] = (
                self.REQUIRED_TAG
                if self.max_workers_required
                else self.max_workers_param.default
            )
        return exec_vals

    @property
    def exec_control_doc(self):
        """str: Execution_control documentation."""
        nodes_doc = NODES_DOC if self.is_split_spatially else ""
        return EXEC_CONTROL_DOC.format(n=nodes_doc, mw=self._max_workers_doc)

    @property
    def _max_workers_doc(self):
        """str: `MW_DOC` formatted with the info from the input func."""
        param = self.param_docs.get("max_workers")
        try:
            return MW_DOC.format(type=param.type, desc=" ".join(param.desc))
        except AttributeError:
            pass

        if self.max_workers_in_func_signature:
            return MW_DOC.format(
                type=(
                    "(int)" if self.max_workers_required else "(int, optional)"
                ),
                desc=(
                    "Maximum number of parallel workers run on each node."
                    "Default is `None`, which uses all available cores."
                ),
            )

        return ""

    @property
    def max_workers_in_func_signature(self):
        """bool: `True` if "max_workers" is a param of the input function."""
        return self.max_workers_param is not None

    @property
    def max_workers_param(self):
        """bool: `True` if "max_workers" is a required parameter of `func`."""
        for sig in self.signatures:
            for name in sig.parameters:
                if name == "max_workers":
                    return sig.parameters["max_workers"]
        return None

    @property
    def max_workers_required(self):
        """bool: `True` if "max_workers" is a required parameter of `func`."""
        param = self.max_workers_param
        if param is None:
            return False
        return param.default is param.empty

    @property
    def required_args(self):
        """set: Required parameters of the input function."""
        required_args = {
            name
            for sig in self.signatures
            for name, param in sig.parameters.items()
            if not name.startswith("_")
            and param.default is param.empty
            and name not in self.skip_params
        }
        return required_args

    @property
    def template_config(self):
        """dict: A template configuration file for this function."""
        config = {"execution_control": self.default_exec_values}
        config.update(
            {
                x: self.REQUIRED_TAG if v.default is v.empty else v.default
                for sig in self.signatures
                for x, v in sig.parameters.items()
                if not x.startswith("_") and x not in self.skip_params
            }
        )
        return config

    @property
    def parameter_help(self):
        """str: Parameter help for the func, including execution control."""
        exec_dict_param = [
            p
            for p in NumpyDocString(self.exec_control_doc)["Parameters"]
            if p.name == "execution_control"
        ]
        param_only_doc = NumpyDocString("")
        param_only_doc["Parameters"] = exec_dict_param + [
            p
            for doc in self.docs
            for p in doc["Parameters"]
            if p.name in self.template_config
        ]
        return str(param_only_doc)

    @property
    def extended_summary(self):
        """str: Function extended summary, with extra whitespace stripped."""
        return "\n".join(
            [x.lstrip().rstrip() for x in self.docs[0]["Extended Summary"]]
        )

    def config_help(self, command_name):
        """Generate a config help string for a command.

        Parameters
        ----------
        command_name : str
            Name of command for which the config help is being
            generated.

        Returns
        -------
        str
            Help string for the config file.
        """
        return CONFIG_DOC.format(
            name=command_name,
            template_json_config=config_as_str_for_docstring(
                self.template_config, config_type=ConfigType.JSON
            ),
            template_yaml_config=config_as_str_for_docstring(
                self.template_config, config_type=ConfigType.YAML
            ),
            template_toml_config=config_as_str_for_docstring(
                self.template_config, config_type=ConfigType.TOML
            ),
            docstring=self.parameter_help,
        )

    def command_help(self, command_name):
        """Generate a help string for a command.

        Parameters
        ----------
        command_name : str
            Name of command for which the help string is being
            generated.

        Returns
        -------
        str
            Help string for the command.
        """
        return COMMAND_DOC.format(
            name=command_name, desc=self.extended_summary
        )


def _pipeline_command_help(pipeline_config):
    """Generate pipeline command help from a sample config."""
    format_inputs = {}
    template_names = [
        "template_json_config",
        "template_yaml_config",
        "template_toml_config",
    ]
    for name, c_type in zip(template_names, CONFIG_TYPES):
        format_inputs[name] = config_as_str_for_docstring(
            pipeline_config, config_type=c_type
        )
    return PIPELINE_CONFIG_DOC.format(**format_inputs)


def _batch_command_help():
    """Generate batch command help from a sample config."""
    template_config = {
        "pipeline_config": FunctionDocumentation.REQUIRED_TAG,
        "sets": [
            {
                "args": FunctionDocumentation.REQUIRED_TAG,
                "files": FunctionDocumentation.REQUIRED_TAG,
                "set_tag": "set1",
            },
            {
                "args": FunctionDocumentation.REQUIRED_TAG,
                "files": FunctionDocumentation.REQUIRED_TAG,
                "set_tag": "set2",
            },
        ],
    }
    sample_args_dict = {
        "args": {
            "input_constant_1": [18.02, 19.04],
            "path_to_a_file": [
                "/first/path.h5",
                "/second/path.h5",
                "/third/path.h5",
            ],
        }
    }
    sample_files = {"files": ["./config_run.yaml", "./config_analyze.json"]}
    format_inputs = {}
    template_names = [
        "template_json_config",
        "template_yaml_config",
        "template_toml_config",
    ]
    for name, c_type in zip(template_names, CONFIG_TYPES):
        format_inputs[name] = config_as_str_for_docstring(
            template_config, config_type=c_type
        )

    sample_args_names = [
        "sample_json_args_dict",
        "sample_yaml_args_dict",
        "sample_toml_args_dict",
    ]
    for name, c_type in zip(sample_args_names, CONFIG_TYPES):
        format_inputs[name] = config_as_str_for_docstring(
            sample_args_dict,
            config_type=c_type,
            num_spaces=20 if "json" in name else 24,
        )
        if "json" in name:
            format_inputs[name] = "\n".join(
                format_inputs[name].split("\n")[1:-1]
            ).lstrip()

    sample_files_names = [
        "sample_json_files",
        "sample_yaml_files",
        "sample_toml_files",
    ]
    for name, c_type in zip(sample_files_names, CONFIG_TYPES):
        format_inputs[name] = config_as_str_for_docstring(
            sample_files,
            config_type=c_type,
            num_spaces=20 if "json" in name else 24,
        )
        if "json" in name:
            format_inputs[name] = "\n".join(
                format_inputs[name].split("\n")[1:-1]
            ).lstrip()

    return BATCH_CONFIG_DOC.format(**format_inputs)



# -*- coding: utf-8 -*-
"""
GAPs CLI template config generation command.
"""
import logging
from functools import partial
from warnings import warn

import click

from rex.utilities.loggers import init_logger
from gaps.config import ConfigType
from gaps.warnings import gapsWarning
from gaps.cli.command import _WrappedCommand


logger = logging.getLogger(__name__)

# pylint: disable=redefined-builtin
@click.pass_context
def _make_template_config(ctx, commands, type, configs):
    """Filter configs and write to file based on type."""
    if ctx.obj.get("VERBOSE"):
        init_logger("gaps")

    configs_to_write = _filter_configs(commands, configs)
    _write_configs(configs_to_write, ConfigType(type))


def _filter_configs(commands, configs):
    """Filter down to only the configs to be written."""
    commands = commands or configs
    configs_to_write = {}
    missing_commands = []
    for command in commands:
        if command not in configs:
            missing_commands.append(command)
            continue
        configs_to_write[command] = configs[command]

    if missing_commands:
        msg = (
            f"The following commands were not found: {missing_commands}. "
            f" Skipping..."
        )
        warn(msg, gapsWarning)

    return configs_to_write


def _write_configs(configs_to_write, config_type):
    """Write out template configs."""
    for command_name, config in configs_to_write.items():
        sample_config_name = f"config_{command_name}.{config_type}"
        sample_config_name = sample_config_name.replace("-", "_")
        logger.info(
            "Generating template config file for command %r: %r",
            command_name,
            sample_config_name,
        )
        if command_name == "pipeline" and "pipeline" in config:
            config["pipeline"] = [
                _update_file_types(pair, config_type)
                for pair in config["pipeline"]
            ]
        config_type.write(sample_config_name, config)


def _update_file_types(pairs, new_type):
    """Update the file endings for all items in the pairs dict."""
    old, new = f".{ConfigType.JSON}", f".{new_type}"
    return {command: path.replace(old, new) for command, path in pairs.items()}


def template_command(template_configs):
    """A template config generation CLI command."""
    allowed_types = " ".join(sorted(f"``{ct}``" for ct in ConfigType))
    params = [
        click.Argument(
            param_decls=["commands"],
            required=False,
            nargs=-1,
        ),
        click.Option(
            param_decls=["--type", "-t"],
            default=f"{ConfigType.JSON}",
            help=f"Configuration file type to generate. Allowed options "
            f"(case-insensitive): {allowed_types}.",
            show_default=True,
        ),
    ]
    return _WrappedCommand(
        "template-configs",
        context_settings=None,
        callback=partial(
            _make_template_config,
            configs=template_configs,
        ),
        params=params,
        help="Generate template config files for requested COMMANDS. If no "
        "COMMANDS are given, config files for the entire pipeline are "
        "generated.",
        epilog=None,
        short_help=None,
        options_metavar="",
        add_help_option=True,
        no_args_is_help=False,
        hidden=False,
        deprecated=False,
    )



# -*- coding: utf-8 -*-
"""
GAPs CLI command for spatially distributed function runs.
"""
import json
import logging
from copy import deepcopy
from warnings import warn
from pathlib import Path
from inspect import signature
from itertools import product

import click

from rex.utilities.loggers import init_mult  # cspell:disable-line

from gaps import ProjectPoints
from gaps.status import StatusUpdates
from gaps.config import load_config
from gaps.log import init_logger
from gaps.cli.execution import kickoff_job
from gaps.exceptions import gapsKeyError
from gaps.warnings import gapsWarning

logger = logging.getLogger(__name__)

_CMD_LIST = [
    "from gaps.cli.config import run_with_status_updates",
    "from {run_func_module} import {run_func_name}",
    "n_args = {node_specific_config}, {logging_options}",
    'su_args = "{project_dir}", "{command}", "{job_name}"',
    "run_with_status_updates("
    "   {run_func_name}, n_args, su_args, {exclude_from_status}"
    ")",
]
TAG = "_j"


class _FromConfig:
    """Utility class for running a function from a config file."""

    def __init__(self, ctx, config_file, command_config):
        """

        Parameters
        ----------
        ctx : click.Context
            Click context for the invoked command.
        config_file : path-like
            Path to input file containing key-value pairs as input to
            function.
        command_config : `gaps.cli.cli.CLICommandConfiguration`
            A command configuration object containing info such as the
            command name, run function, pre-processing function,
            function documentation, etc.
        """
        self.ctx = ctx
        self.config_file = Path(config_file).expanduser().resolve()
        self.command_config = command_config
        self.config = load_config(config_file)
        self.log_directory = None
        self.verbose = None
        self.exec_kwargs = None
        self.logging_options = None
        self.exclude_from_status = None

    @property
    def project_dir(self):
        """`Path`: Path to project directory."""
        return self.config_file.parent

    @property
    def command_name(self):
        """str: Name of command being run."""
        return self.command_config.name

    @property
    def job_name(self):
        """str: Name of job being run."""
        return "_".join(
            [self.project_dir.name, self.command_name.replace("-", "_")]
        )

    def enable_logging(self):
        """Enable logging based on config file input."""
        self.log_directory = self.config.pop(
            "log_directory", (self.project_dir / "logs").as_posix()
        )
        self.verbose = (
            self.config.pop("log_level", "INFO") == "DEBUG"
            or self.ctx.obj["VERBOSE"]
        )
        pipe_log_file = Path(self.log_directory) / f"{self.job_name}.log"
        init_logger(
            stream=self.ctx.obj.get("LOG_STREAM", True),
            level="DEBUG" if self.verbose else "INFO",
            file=pipe_log_file.as_posix(),
        )
        return self

    def validate_config(self):
        """Validate the user input config file."""
        logger.debug("Validating %r", self.config_file)
        _validate_config(
            self.config, self.command_config.function_documentation
        )
        return self

    def preprocess_config(self):
        """Apply preprocessing function to config file."""

        preprocessor_kwargs = {
            "config": self.config,
            "command_name": self.command_name,
            "config_file": self.config_file,
            "project_dir": self.project_dir,
            "job_name": self.job_name,
            "out_dir": self.project_dir,
        }
        extra_preprocessor_kwargs = {
            k: self.config[k]
            for k in self.command_config.preprocessor_args
            if k not in preprocessor_kwargs and k in self.config
        }
        preprocessor_kwargs.update(extra_preprocessor_kwargs)
        preprocessor_defaults = {
            k: v
            for k, v in self.command_config.preprocessor_defaults.items()
            if k not in preprocessor_kwargs
        }
        preprocessor_kwargs.update(preprocessor_defaults)
        preprocessor_kwargs = {
            k: v
            for k, v in preprocessor_kwargs.items()
            if k in self.command_config.preprocessor_args
        }
        self.config = self.command_config.config_preprocessor(
            **preprocessor_kwargs
        )
        return self

    def set_exec_kwargs(self):
        """Extract the execution control dictionary."""
        self.exec_kwargs = {
            "option": "local",
            "sh_script": "",
            "stdout_path": (Path(self.log_directory) / "stdout").as_posix(),
        }

        self.exec_kwargs.update(self.config.get("execution_control", {}))
        if "max_workers" in self.config:
            max_workers = self.config.pop("max_workers")
            msg = (
                "Found key 'max_workers' outside of 'execution_control'. "
                f"Moving 'max_workers' value ({max_workers}) into "
                "'execution_control' block. To silence this warning, "
                "please specify 'max_workers' inside of the "
                "'execution_control' block."
            )
            warn(msg, gapsWarning)
            self.exec_kwargs["max_workers"] = max_workers

        return self

    def set_logging_options(self):
        """Assemble the logging options dictionary."""
        self.logging_options = {
            "name": self.job_name,
            "log_directory": self.log_directory,
            "verbose": self.verbose,
            "node": self.exec_kwargs.get("option", "local") != "local",
        }
        return self

    def set_exclude_from_status(self):
        """Assemble the exclusion keyword set."""
        self.exclude_from_status = {"project_points"}
        self.exclude_from_status |= set(
            self.config.pop("exclude_from_status", set())
        )
        self.exclude_from_status = [
            key.lower()
            for key in self.exclude_from_status
            if key.lower() != "tag"
        ]
        return self

    def prepare_context(self):
        """Add required key-val;ue pairs to context object."""
        self.ctx.obj["COMMAND_NAME"] = self.command_name
        self.ctx.obj["OUT_DIR"] = self.project_dir
        return self

    def kickoff_jobs(self):
        """Kickoff jobs across nodes based on config and run function."""
        keys_to_run, lists_to_run = self._keys_and_lists_to_run()

        jobs = sorted(product(*lists_to_run))
        num_jobs_submit = len(jobs)
        n_zfill = len(str(num_jobs_submit))
        max_workers_per_node = self.exec_kwargs.pop("max_workers", None)
        for node_index, values in enumerate(jobs):
            if num_jobs_submit > 1:
                tag = f"{TAG}{str(node_index).zfill(n_zfill)}"
            else:
                tag = ""
            self.ctx.obj["NAME"] = job_name = f"{self.job_name}{tag}"
            node_specific_config = deepcopy(self.config)
            node_specific_config.pop("execution_control", None)
            node_specific_config.update(
                {
                    "tag": tag,
                    "command_name": self.command_name,
                    "config_file": self.config_file.as_posix(),
                    "project_dir": self.project_dir.as_posix(),
                    "job_name": job_name,
                    "out_dir": self.project_dir.as_posix(),
                    "max_workers": max_workers_per_node,
                }
            )

            for key, val in zip(keys_to_run, values):
                if isinstance(key, str):
                    node_specific_config.update({key: val})
                else:
                    node_specific_config.update(dict(zip(key, val)))

            cmd = "; ".join(_CMD_LIST).format(
                run_func_module=self.command_config.function.__module__,
                run_func_name=self.command_config.function.__name__,
                node_specific_config=as_script_str(node_specific_config),
                project_dir=self.project_dir.as_posix(),
                logging_options=as_script_str(self.logging_options),
                exclude_from_status=as_script_str(self.exclude_from_status),
                command=self.command_name,
                job_name=job_name,
            )
            cmd = f"python -c {cmd!r}"
            kickoff_job(self.ctx, cmd, deepcopy(self.exec_kwargs))

        return self

    def _keys_and_lists_to_run(self):
        """Compile run lists based on `command_config.split_keys` input."""
        keys_to_run = []
        lists_to_run = []
        keys = sorted(self.command_config.split_keys, key=_project_points_last)
        for key_group in keys:
            keys_to_run.append(key_group)
            if isinstance(key_group, str):
                lists_to_run.append(self.config.get(key_group) or [None])
            else:
                lists_to_run.append(
                    list(
                        zip(*[self.config.get(k) or [None] for k in key_group])
                    )
                )
        return keys_to_run, lists_to_run

    def run(self):
        """Run the entire config pipeline."""
        return (
            self.enable_logging()
            .validate_config()
            .preprocess_config()
            .set_exec_kwargs()
            .set_logging_options()
            .set_exclude_from_status()
            .prepare_context()
            .kickoff_jobs()
        )


@click.pass_context
def from_config(ctx, config_file, command_config):
    """Run command from a config file."""
    _FromConfig(ctx, config_file, command_config).run()


def _validate_config(config, function_documentation):
    """Ensure required keys exist and warn user about extra keys in config."""
    _ensure_required_args_exist(config, function_documentation)
    _warn_about_extra_args(config, function_documentation)


def _ensure_required_args_exist(config, function_documentation):
    """Make sure that args required for func to run exist in config."""
    missing = {
        name
        for name in function_documentation.required_args
        if name not in config
    }

    if function_documentation.max_workers_required:
        missing = _mark_max_workers_missing_if_needed(config, missing)

    if any(missing):
        msg = (
            f"The following required keys are missing from the configuration "
            f"file: {missing}"
        )
        raise gapsKeyError(msg)


def _mark_max_workers_missing_if_needed(config, missing):
    """Add max_workers as missing key if it is not in `execution_control."""

    exec_control = config.get("execution_control", {})
    if "max_workers" not in config and "max_workers" not in exec_control:
        missing |= {"max_workers"}
    return missing


def _warn_about_extra_args(config, function_documentation):
    """Warn user about extra unused keys in the config file."""
    extra = {
        name
        for name in config.keys()
        if not _param_in_sig(name, function_documentation)
    }
    extra -= {"execution_control", "project_points_split_range"}
    if any(extra):
        msg = (
            "Found unused keys in the configuration file: %s. To silence "
            "this warning, please remove these keys from the input "
            "configuration file."
        )
        warn(msg % extra, gapsWarning)


def _param_in_sig(param, function_documentation):
    """Determine if ``name`` is an argument in any func signatures"""
    return any(
        param in _public_args(signature)
        for signature in function_documentation.signatures
    )


def _public_args(func_signature):
    """Gather set of all "public" function args."""
    return {
        param
        for param in func_signature.parameters.keys()
        if not param.startswith("_")
    }


def _project_points_last(key):
    """Sorting function that always puts "project_points_split_range" last."""
    if isinstance(key, str):
        if key.casefold() == "project_points_split_range":
            return (chr(0x10FFFF),)  # PEP 393
        return (key,)
    return key


def as_script_str(input_):
    """Convert input to how it would appear in a python script.

    Essentially this means the input is dumped to json format with some
    minor replacements (e.g. null -> None, etc.). Importantly, all
    string inputs are wrapped in double quotes.

    Parameters
    ----------
    input_ : obj
        Any object that can be serialized with :func:`json.dumps`.

    Returns
    -------
    str
        Input object as it would appear in a python script.

    Examples
    --------
    >>> as_script_str("a")
    "a"

    >>> as_script_str({"a": None, "b": True, "c": False, "d": 3,
    ...                "e": [{"t": "hi"}]})
    {"a": None, "b": True, "c": False, "d": 3, "e": [{"t": "hi"}]}
    """
    return (
        json.dumps(input_)
        .replace("null", "None")
        .replace("false", "False")
        .replace("true", "True")
    )


def run_with_status_updates(run_func, node_args, status_update_args, exclude):
    """Run a function and write status updated before/after execution.

    Parameters
    ----------
    run_func : callable
        A function to run.
    node_args : iterable
        An iterable containing the last three arguments to
        :func:`node_kwargs`.
    status_update_args : iterable
        An iterable containing the first three initializer arguments for
        :class:`StatusUpdates`.
    exclude : collection | None
        A collection (list, set, dict, etc.) of keys that should be
        excluded from the job status file that is written before/after
        the function runs.
    """
    run_kwargs = node_kwargs(run_func, *node_args)
    exclude = exclude or set()
    job_attrs = {
        key: value for key, value in run_kwargs.items() if key not in exclude
    }
    status_update_args = *status_update_args, job_attrs
    with StatusUpdates(*status_update_args) as status:
        status.out_file = run_func(**run_kwargs)


def node_kwargs(run_func, config, logging_options):
    """Compile the function inputs arguments for a particular node.

    Parameters
    ----------
    run_func : callable
        A function to run.
    config : dict
        Dictionary of node-specific inputs to `run_func`.
    logging_options : dict
        Dictionary of logging options containing at least the following
        key-value pairs:

            name : str
                Job name; name of log file.
            log_directory : path-like
                Path to log file directory.
            verbose : bool
                Option to turn on debug logging.
            node : bool
                Flag for whether this is a node-level logger. If this is
                a node logger, and the log level is info, the log_file
                will be `None` (sent to stdout).


    Returns
    -------
    dict
        Function run kwargs to be used on this node.
    """

    # initialize loggers for multiple modules
    init_mult(  # cspell:disable-line
        logging_options["name"],
        logging_options["log_directory"],
        modules=[run_func.__module__.split(".")[0], "gaps", "rex"],
        verbose=logging_options["verbose"],
        node=logging_options["node"],
    )

    split_range = config.pop("project_points_split_range", None)
    if split_range is not None:
        config["project_points"] = ProjectPoints.from_range(
            split_range, config["project_points"]
        )

    sig = signature(run_func)
    run_kwargs = {
        k: v for k, v in config.items() if k in sig.parameters.keys()
    }
    logger.debug("Running %r with kwargs: %s", run_func.__name__, run_kwargs)
    return run_kwargs



# -*- coding: utf-8 -*-
"""
GAPs Pipeline architecture.
"""
import time
import logging
from pathlib import Path
from warnings import warn

from gaps.hpc import SLURM
from gaps.status import Status, StatusOption, StatusField
from gaps.utilities import recursively_update_dict
from gaps.config import load_config, init_logging_from_config
from gaps.exceptions import (
    gapsConfigError,
    gapsExecutionError,
    gapsKeyError,
    gapsValueError,
)
from gaps.warnings import gapsWarning


logger = logging.getLogger(__name__)


class Pipeline:
    """gaps pipeline execution framework."""

    COMMANDS = {}

    def __init__(self, pipeline, monitor=True):
        """

        Parameters
        ----------
        pipeline : path-like
            Pipeline config file path.
        monitor : bool, optional
            Flag to perform continuous monitoring of the pipeline.
            By default, ``True``.

        Raises
        ------
        gapsConfigError
            If "pipeline" key not in config file.
        """
        self.monitor = monitor
        self._out_dir = Path(pipeline).expanduser().parent.resolve()
        self._out_dir = self._out_dir.as_posix()

        config = load_config(pipeline)

        if "pipeline" not in config:
            raise gapsConfigError(
                'Could not find required key "pipeline" in the '
                "pipeline config."
            )
        self._run_list = config["pipeline"]
        self._init_status()

        init_logging_from_config(config)

    def _init_status(self):
        """Initialize the status json in the output directory."""
        status = self.status
        for pipe_index, step in enumerate(self._run_list):
            for command in step.keys():
                command_dict = {
                    command: {StatusField.PIPELINE_INDEX: pipe_index}
                }
                status.data = recursively_update_dict(
                    status.data, command_dict
                )

        _dump_sorted(status)

    @property
    def status(self):
        """:class:`~gaps.status.Status`: A gaps pipeline status object."""
        return Status(self._out_dir).update_from_all_job_files(purge=False)

    @property
    def name(self):
        """str: Name of the pipeline job (directory of status file)."""
        return self.status.name

    def _cancel_all_jobs(self):
        """Cancel all jobs in this pipeline."""
        slurm_manager = SLURM()
        for job_id in self.status.job_ids:
            slurm_manager.cancel(job_id)
        logger.info("Pipeline job %r cancelled.", self.name)

    def _main(self):
        """Iterate through run list submitting steps while monitoring status"""

        for step, command in enumerate(self._run_list):
            step_status = self._status(step)

            if step_status == StatusOption.SUCCESSFUL:
                logger.debug("Successful: %r.", list(command.keys())[0])
                continue

            # the submit function handles individual job success/failure
            self._submit(step)

            # do not enter while loop for continuous monitoring
            if not self.monitor:
                break

            time.sleep(1)
            self._monitor(step)

        else:
            logger.info("Pipeline job %r is complete.", self.name)
            logger.debug("Output directory is: %r", self._out_dir)

    def _monitor(self, step, seconds=5, step_status=StatusOption.RUNNING):
        """Continuously monitor job until success or failure."""

        while step_status.is_processing:
            time.sleep(seconds)
            step_status = self._status(step)

            if step_status == StatusOption.FAILED:
                command, f_config = self._get_command_config(step)
                raise gapsExecutionError(
                    f"Pipeline failed at step {step}: {command!r} "
                    f"for {f_config!r}"
                )

    def _submit(self, step):
        """Submit a step in the pipeline."""

        command, f_config = self._get_command_config(step)
        if command not in self.COMMANDS:
            raise gapsKeyError(
                f"Could not recognize command {command!r}. "
                f"Available commands are: {set(self.COMMANDS)!r}"
            ) from None

        self.COMMANDS[command].callback(f_config)

    def _status(self, step):
        """Get a pipeline step status."""

        command, _ = self._get_command_config(step)
        status = self.status
        submitted = _check_jobs_submitted(status, command)
        if not submitted:
            return StatusOption.RUNNING

        return self._get_command_return_code(status, command)

    def _get_command_return_code(self, status, command):
        """Get a return code for a command based on a status object.

        Note that it is assumed a job has been submitted before this
        function is called, otherwise the return values make no sense!
        """

        # initialize return code array
        arr = []
        check_failed = False
        status.update_from_all_job_files()

        if command not in status.data:
            # assume running
            arr = [StatusOption.RUNNING]
        else:
            for job_name, job_info in status.data[command].items():
                if job_name == StatusField.PIPELINE_INDEX:
                    continue

                job_status = job_info[StatusField.JOB_STATUS]

                if job_status == "successful":
                    arr.append(StatusOption.SUCCESSFUL)
                elif job_status == "failed":
                    arr.append(StatusOption.FAILED)
                    check_failed = True
                elif job_status == "submitted":
                    arr.append(StatusOption.SUBMITTED)
                elif job_status == "running":
                    arr.append(StatusOption.RUNNING)
                elif job_status is None:
                    arr.append(StatusOption.COMPLETE)
                else:
                    msg = "Job status code {job_status!r} not understood!"
                    raise gapsValueError(msg)

            _dump_sorted(status)

        return_code = _parse_code_array(arr)

        fail_str = ""
        if return_code != StatusOption.FAILED and check_failed:
            fail_str = ", but some jobs have failed"
        logger.info(
            "CLI command %r for job %r %s%s.",
            command,
            self.name,
            return_code.with_verb,  # pylint: disable=no-member
            fail_str,
        )

        return return_code

    def _get_command_config(self, step):
        """Get the (command, config) key pair."""
        return list(self._run_list[step].items())[0]

    @classmethod
    def cancel_all(cls, pipeline):
        """Cancel all jobs corresponding to pipeline.

        Parameters
        ----------
        pipeline : path-like
            Pipeline config file path.
        """
        cls(pipeline)._cancel_all_jobs()

    @classmethod
    def run(cls, pipeline, monitor=True):
        """Run the pipeline.

        Parameters
        ----------
        pipeline : path-like
            Pipeline config file path.
        monitor : bool
            Flag to perform continuous monitoring of the pipeline.
        """
        cls(pipeline, monitor=monitor)._main()


def _parse_code_array(arr):
    """Parse array of return codes to get single return code for command."""

    # check to see if all have completed, or any have failed
    all_successful = all(status == StatusOption.SUCCESSFUL for status in arr)
    all_completed = not any(status.is_processing for status in arr)
    any_failed = any(status == StatusOption.FAILED for status in arr)
    any_submitted = any(status == StatusOption.SUBMITTED for status in arr)

    # only return success if all have succeeded.
    if all_successful:
        return StatusOption.SUCCESSFUL
    # Only return failed when all have finished.
    if all_completed and any_failed:
        return StatusOption.FAILED
    # only return complete when all have completed
    # (but some should have succeeded or failed)
    if all_completed:
        return StatusOption.COMPLETE
    # only return "running" if all jobs running, else return "submitted"
    if any_submitted:
        return StatusOption.SUBMITTED
    # otherwise, all jobs are still running
    return StatusOption.RUNNING


def _check_jobs_submitted(status, command):
    """Check whether jobs have been submitted for a given command."""
    return any(
        job != StatusField.PIPELINE_INDEX for job in status.data.get(command)
    )


def _dump_sorted(status):
    """Dump status dict after sorting on PIPELINE_INDEX."""
    pi_key = StatusField.PIPELINE_INDEX

    def _sort_key(status_entry):
        """Sort on pipeline index and key name, putting non-pipeline at top."""
        try:
            pipe_index = status[status_entry].get(pi_key, -1)
        except AttributeError:
            pipe_index = -1
        return pipe_index, status_entry

    sorted_keys = sorted(status, key=_sort_key)
    status.data = {k: status[k] for k in sorted_keys}
    status.dump()


def parse_previous_status(status_dir, command, key=StatusField.OUT_FILE):
    """Parse key from job status(es) from the previous pipeline step.

    Parameters
    ----------
    status_dir : path-like
        Directory containing the status file to parse.
    command : str
        Current CLI command (i.e. current pipeline step).
    key : `StatusField` | str, optional
        Parsing target of previous command. By default,
        `StatusField.OUT_FILE`.

    Returns
    -------
    out : list
        Arguments parsed from the status file in status_dir from
        the command preceding the input command arg. This list is
        empty if the `key` is not found in the job status, or if
        the previous step index does not exist in status.

    Raises
    ------
    gapsKeyError
        If ``command`` not in status.
    """

    status = Status(status_dir).update_from_all_job_files(purge=False)

    if (
        command not in status
        or StatusField.PIPELINE_INDEX not in status[command]
    ):
        msg = (
            f"Could not parse data for command {command!r} from status "
            f"file in {status_dir!r}"
        )
        raise gapsKeyError(msg)

    index = int(status.data[command][StatusField.PIPELINE_INDEX]) - 1

    if index < 0:
        index = 0
        msg = (
            f"CLI command {command!r} is attempting to parse a previous "
            f"pipeline step, but it appears to be the first step. "
            f"Attempting to parse data from {command!r}."
        )
        warn(msg, gapsWarning)

    for cmd, status in status.items():
        try:
            command_index = status.get(StatusField.PIPELINE_INDEX)
        except AttributeError:
            continue

        if str(index) == str(command_index):
            return Status.parse_command_status(status_dir, cmd, key)

    return []



# -*- coding: utf-8 -*-
# pylint: disable=invalid-name
"""Custom Warning for GAPs. """
import logging

logger = logging.getLogger("gaps")


class gapsWarning(UserWarning):
    """Generic gaps Warning."""

    def __init__(self, *args, **kwargs):
        """Init exception and broadcast message to logger."""
        super().__init__(*args, **kwargs)
        if args:
            logger.warning(str(args[0]), stacklevel=2)


class gapsCollectionWarning(gapsWarning):
    """gaps Collection waring."""


class gapsHPCWarning(gapsWarning):
    """gaps HPC warning."""



# -*- coding: utf-8 -*-
"""
GAPs execution CLI utils.
"""
import logging
import datetime as dt
from warnings import warn
from inspect import signature

from gaps.hpc import submit
from gaps.status import (
    DT_FMT,
    Status,
    HardwareOption,
    StatusOption,
    StatusField,
)
from gaps.warnings import gapsWarning
from gaps.exceptions import gapsConfigError

logger = logging.getLogger(__name__)


def kickoff_job(ctx, cmd, exec_kwargs):
    """Kickoff a single job (a single command execution).

    Parameters
    ----------
    ctx : click.Context
        Context object with a `.obj` attribute that contains at least
        the following keys:

            NAME : str
                Job name.
            OUT_DIR : path-like
                Path to output directory.
            COMMAND_NAME : str
                Name of command being run.

    cmd : str
        String form of command to kickoff.
    exec_kwargs : dict
        Keyword-value pairs to pass to the respective `submit` function.
        These will be filtered, so they may contain extra values. If
        some required inputs are missing from this dictionary, a
        `gapsConfigError` is raised.

    Raises
    ------
    gapsConfigError
        If `exec_kwargs` is missing some arguments required by the
        respective `submit` function.
    """
    hardware_option = HardwareOption(exec_kwargs.pop("option", "local"))
    if hardware_option.manager is None:
        _kickoff_local_job(ctx, cmd)
        return

    ctx.obj["MANAGER"] = hardware_option.manager
    exec_kwargs = _filter_exec_kwargs(
        exec_kwargs, hardware_option.manager.make_script_str
    )
    _kickoff_hpc_job(ctx, cmd, **exec_kwargs)


def _filter_exec_kwargs(kwargs, func):
    """Filter out extra keywords and raise error if any are missing."""
    sig = signature(func)
    kwargs_to_use = {
        k: v for k, v in kwargs.items() if k in sig.parameters.keys()
    }
    extra_keys = set(kwargs) - set(kwargs_to_use)
    if extra_keys:
        msg = (
            f"Found extra keys in 'execution_control'! The following "
            f"inputs will be ignored: {extra_keys}"
        )
        warn(msg, gapsWarning)

    required = {
        name for name, p in sig.parameters.items() if p.default == p.empty
    }
    required -= {"self", "cmd", "name"}
    missing = {k for k in required if k not in kwargs_to_use}
    if missing:
        msg = (
            f"The 'execution_control' block is missing the following "
            f"required keys: {missing}"
        )
        raise gapsConfigError(msg)

    return kwargs_to_use


def _kickoff_local_job(ctx, cmd):
    """Run a job (command) locally."""

    if not _should_run(ctx):
        return

    name = ctx.obj["NAME"]
    logger.info("Running locally with job name %r.", name)
    logger.debug("Submitting the following command:\n%s", cmd)
    Status.mark_job_as_submitted(
        ctx.obj["OUT_DIR"],
        command=ctx.obj["COMMAND_NAME"],
        job_name=name,
        replace=True,
        job_attrs={
            StatusField.JOB_STATUS: StatusOption.SUBMITTED,
            StatusField.HARDWARE: HardwareOption.LOCAL,
            StatusField.TIME_SUBMITTED: dt.datetime.now().strftime(DT_FMT),
        },
    )
    stdout, stderr = submit(cmd)
    if stdout:
        logger.info("Subprocess received stdout: \n%s", stdout)
    if stderr:
        logger.warning("Subprocess received stderr: \n%s", stderr)
    msg = f"Completed job {name!r}."
    logger.info(msg)


def _kickoff_hpc_job(ctx, cmd, **kwargs):
    """Run a job (command) on the HPC."""

    if not _should_run(ctx):
        return

    name = ctx.obj["NAME"]
    logger.info("Running on HPC with job name %r.", name)
    logger.debug("Submitting the following command:\n%s", cmd)
    out = ctx.obj["MANAGER"].submit(name, cmd=cmd, **kwargs)[0]
    id_msg = f" (Job ID #{out})" if out else ""
    msg = f"Kicked off job {name!r}{id_msg}"

    Status.mark_job_as_submitted(
        ctx.obj["OUT_DIR"],
        command=ctx.obj["COMMAND_NAME"],
        job_name=name,
        replace=True,
        job_attrs={
            StatusField.JOB_ID: out,
            StatusField.HARDWARE: HardwareOption.SLURM,
            StatusField.JOB_STATUS: StatusOption.SUBMITTED,
            StatusField.TIME_SUBMITTED: dt.datetime.now().strftime(DT_FMT),
        },
    )
    logger.info(msg)


def _should_run(ctx):
    """Determine wether a command should be run based on status."""
    name = ctx.obj["NAME"]
    out_dir = ctx.obj["OUT_DIR"]
    status = Status.retrieve_job_status(
        out_dir,
        command=ctx.obj["COMMAND_NAME"],
        job_name=name,
        subprocess_manager=ctx.obj.get("MANAGER"),
    )
    if status == StatusOption.SUCCESSFUL:
        msg = (
            f"Job {name!r} is successful in status json found in {out_dir!r}, "
            f"not re-running."
        )
        logger.info(msg)
        return False

    if status is not None and "fail" not in str(status).lower():
        msg = (
            f"Job {name!r} was found with status {status!r}, not resubmitting"
        )
        logger.info(msg)
        return False

    return True



# -*- coding: utf-8 -*-
# pylint: disable=too-few-public-methods
"""
GAPs command configuration preprocessing functions.
"""
from inspect import signature

import click

from gaps.cli.documentation import FunctionDocumentation
from gaps.cli.preprocessing import split_project_points_into_ranges


GAPS_SUPPLIED_ARGS = {
    "tag",
    "command_name",
    "config_file",
    "project_dir",
    "job_name",
    "out_dir",
    "config",
}


class CLICommandConfiguration:
    """Configure a CLI command to execute a function on multiple nodes.

    This class configures a CLI command that runs a given function
    across multiple nodes on an HPC. The primary utility is to split the
    function execution spatially, meaning that individual nodes will run
    the function on a subset of the input points. However, this
    configuration also supports splitting the execution on other inputs
    (in lieu of or in addition to the geospatial partitioning).
    """

    def __init__(
        self, name, function, split_keys=None, config_preprocessor=None
    ):
        """

        Parameters
        ----------
        name : str
            Name of the command. This will be the name used to call the
            command on the terminal. This name does not have to match
            the function name. It is encouraged to use lowercase names
            with dashes ("-") instead of underscores ("_") to stay
            consistent with click's naming conventions.
        function : callable
            The function to run on individual nodes. This function will
            be used to generate all the documentation and template
            configuration files, so it should be thoroughly documented
            (using a `NumPy Style Python Docstring
            <https://numpydoc.readthedocs.io/en/latest/format.html>`_).
            This function must return the path to the output file it
            generates (or a list of paths if multiple output files are
            generated). If no output files are generated, the function
            must return ``None`` or an empty list. In order to avoid
            clashing output file names with jobs on other nodes, make
            sure to "request" the ``tag`` argument. This function can
            "request" the following arguments by including them in the
            function signature (``gaps`` will automatically pass them to
            the function without any additional used input):

                tag : str
                    Short string unique to this job run that can be used
                    to generate unique output filenames, thereby
                    avoiding clashing output files with jobs on other
                    nodes. This string  contains a leading underscore,
                    so the file name can easily be generated:
                    ``f"{out_file_name}{tag}.{extension}"``.
                    See :func:`gaps.cli.collect.collect` for an example.
                command_name : str
                    Name of the command being run. This is equivalent to
                    the ``name`` input above.
                config_file : Path
                    Path to the configuration file specified by the
                    user.
                project_dir : Path
                    Path to the project directory (parent directory of
                    the configuration file).
                job_name : str
                    Name of the job being run. This is typically a
                    combination of the project directory and the command
                    name.
                out_dir : Path
                    Path to output directory - typically equivalent to
                    the project directory.

            If your function is capable of multiprocessing, you should
            also include ``max_workers`` in the function signature.
            ``gaps`` will pass an integer equal to the number of
            processes the user wants to run on a single node for this
            value. Note that the ``config`` parameter is not allowed as
            a function signature item. Please request all the required
            keys/inputs directly. This function can also request
            "private" arguments by including a leading underscore in the
            argument name. These arguments are NOT exposed to users in
            the documentation or template configuration files. Instead,
            it is expected that the ``config_preprocessor`` function
            fills these arguments in programmatically before the
            function is distributed across nodes. See the implementation
            of :func:`gaps.cli.preprocessing.preprocess_collect_config`
            and :func:`gaps.cli.collect.collect` for an example of this
            pattern.
        split_keys : set | container, optional
            A set of strings identifying the names of the config keys
            that ``gaps`` should split the function execution on. To
            specify geospatial partitioning in particular, ensure that
            the main ``function`` has a "project_points" argument (which
            accepts a :class:`gaps.project_points.ProjectPoints`
            instance) and specify "project_points" as a split argument.
            Users of the CLI will only need to specify the path to the
            project points file and a "nodes" argument in the execution
            control. To split execution on additional/other inputs,
            include them by name in this input (and ensure the run
            function accepts them as input). You may include tuples of
            strings in this iterable as well. Tuples of strings will be
            interpreted as combinations of keys whose values should be
            iterated over simultaneously. For example, specifying
            ``split_keys=[("a", "b")]`` and invoking with a config file
            where ``a = [1, 2]`` and ``b = [3, 4]`` will run the main
            function two times (on two nodes), first with the inputs
            ``a=1, b=3`` and then with the inputs ``a=2, b=4``.
            It is the responsibility of the developer using this class
            to ensure that the user input for all ``split_keys`` is an
            iterable (typically a list), and that the lengths of all
            "paired" keys match. To allow non-iterable user input for
            split keys, use the ``config_preprocessor`` argument to
            specify a preprocessing function that converts the user
            input into a list of the expected inputs. If ``None``,
            execution is not split across nodes, and a single node is
            always used for the function call. By default, ``None``.
        config_preprocessor : callable, optional
            Optional function for configuration pre-processing. At
            minimum, this function should have "config" as an argument,
            which will receive the user configuration input as a
            dictionary. This function can also "request" the following
            arguments by including them in the function signature:

                command_name : str
                    Name of the command being run. This is equivalent to
                    the ``name`` input above.
                config_file : Path
                    Path to the configuration file specified by the
                    user.
                project_dir : Path
                    Path to the project directory (parent directory of
                    the configuration file).
                job_name : str
                    Name of the job being run. This is typically a
                    combination of the project directory and the command
                    name.
                out_dir : Path
                    Path to output directory - typically equivalent to
                    the project directory.

            See :func:`gaps.cli.preprocessing.preprocess_collect_config`
            for an example. Note that the ``tag`` parameter is not
            allowed as a pre-processing function signature item (the
            node jobs will not have been configured before this function
            executes). This function can also "request" new user inputs
            that are not present in the signature of the main run
            function. In this case, the documentation for these new
            arguments is pulled from the ``config_preprocessor``
            function. This feature can be used to request auxillary
            information from the user to fill in "private" inputs to the
            main run function. See the implementation of
            :func:`gaps.cli.preprocessing.preprocess_collect_config` and
            :func:`gaps.cli.collect.collect` for an example of this
            pattern. By default, ``None``.
        """
        self.name = name
        self.function = function
        self.split_keys = set() if split_keys is None else set(split_keys)
        self.config_preprocessor = config_preprocessor or _passthrough
        self.function_documentation = FunctionDocumentation(
            self.function,
            self.config_preprocessor,
            skip_params=GAPS_SUPPLIED_ARGS,
            is_split_spatially=self.is_split_spatially,
        )
        preprocessor_sig = signature(self.config_preprocessor)
        self.preprocessor_args = preprocessor_sig.parameters.keys()
        self.preprocessor_defaults = {
            name: param.default
            for name, param in preprocessor_sig.parameters.items()
            if param.default != param.empty
        }
        if self.is_split_spatially:
            self._add_split_on_points()

    def _add_split_on_points(self):
        """Add split points preprocessing."""
        self.config_preprocessor = _split_points(self.config_preprocessor)
        self.split_keys -= {"project_points"}
        self.split_keys |= {"project_points_split_range"}

    @property
    def is_split_spatially(self):
        """bool: ``True`` if execution is split spatially across nodes."""
        return any(
            key in self.split_keys
            for key in ["project_points", "project_points_split_range"]
        )


def _passthrough(config):
    """Pass the input config through with no modifications."""
    return config


def _split_points(config_preprocessor):
    """Add the `split_project_points_into_ranges` to preprocessing."""

    def _config_preprocessor(config, *args, **kwargs):
        config = config_preprocessor(config, *args, **kwargs)
        config = split_project_points_into_ranges(config)
        return config

    return _config_preprocessor


# pylint: disable=invalid-name,unused-argument
class _WrappedCommand(click.Command):
    """Click Command class with an updated `get_help` function.

    References
    ----------
    https://stackoverflow.com/questions/55585564/python-click-formatting-help-text
    """

    _WRAP_TEXT_REPLACED = False

    def get_help(self, ctx):
        """Format the help into a string and return it."""
        if self._WRAP_TEXT_REPLACED:
            return super().get_help(ctx)

        orig_wrap_test = click.formatting.wrap_text

        def wrap_text(
            text,
            width=78,
            initial_indent="",
            subsequent_indent="",
            preserve_paragraphs=False,
        ):
            """Wrap text with gaps-style newline handling."""
            wrapped_text = orig_wrap_test(
                text.replace("\n", "\n\n"),
                width,
                initial_indent=initial_indent,
                subsequent_indent=subsequent_indent,
                preserve_paragraphs=True,
            )
            wrapped_text = (
                wrapped_text.replace("\n\n", "\n").replace("::\n", ":\n\n")
                # .replace("}\nParameters", "}\n\nParameters")
                .replace("[required]", "\n[required]")
            )
            if "Parameters\n----------" not in wrapped_text.replace(" ", ""):
                wrapped_text = wrapped_text.replace(".\n", ".\n\n")
            return wrapped_text

        click.formatting.wrap_text = wrap_text
        self._WRAP_TEXT_REPLACED = True
        return super().get_help(ctx)



# -*- coding: utf-8 -*-
"""
GAPs Project Points
"""
import logging
from warnings import warn

import numpy as np
import pandas as pd

from rex.utilities import parse_table
from gaps.exceptions import (
    gapsIndexError,
    gapsKeyError,
    gapsRuntimeError,
)
from gaps.utilities import project_points_from_container_or_slice


logger = logging.getLogger(__name__)


def _parse_sites(points):
    """Parse project points from list or slice

    Parameters
    ----------
    points : int | str | pd.DataFrame | slice | list
        Slice specifying project points, string pointing to a project
        points csv, or a DataFrame containing the effective csv contents.
        Can also be a single integer site value.

    Returns
    -------
    df : pd.DataFrame
        DataFrame of sites (gids) with corresponding args

    Raises
    ------
    gapsRuntimeError
        If points not flat.
    """
    try:
        points = project_points_from_container_or_slice(points)
    except TypeError as err:
        msg = (
            f"Cannot parse points data from {points}. If this input is a "
            "container, please ensure that the container is flat (no "
            "nested gid values)."
        )
        raise gapsRuntimeError(msg) from err

    return pd.DataFrame({"gid": points})


class ProjectPoints:
    """Class to manage site and SAM input configuration requests."""

    def __init__(self, points, **kwargs):
        """Initialize ProjectPoints.

        Parameters
        ----------
        points : int | slice | list | tuple | str | pd.DataFrame | dict
            Slice specifying project points, string pointing to a
            project points csv, or a DataFrame containing the effective
            csv contents. Can also be a single integer site value.
        **kwargs
            Keyword-argument pairs to add to project points DataFrame.
            The key should be the column name, and the value should
            be the value to add under the column name. Values must
            either be a scalar or match the length of the DataFrame
            resulting from the `points` input.
        """
        self._df = self.split_range = None
        self._parse_points(points, **kwargs)

    def _parse_points(self, points, **kwargs):
        """Generate the project points df from inputs

        Parameters
        ----------
        points : int | str | pd.DataFrame | slice | list | dict
            Slice specifying project points, string pointing to a
            project points csv, or a DataFrame containing the effective
            csv contents. Can also be a single integer site value.

        Returns
        -------
        df : pd.DataFrame
            DataFrame of sites (gids) with corresponding args
        """
        try:
            self._df = parse_table(points)
        except ValueError:
            self._df = _parse_sites(points)

        if "gid" not in self._df.columns:
            raise gapsKeyError(
                'Project points data must contain "gid" column.'
            )

        for key, val in kwargs.items():
            self._df[key] = val

        self.split_range = (self._df.iloc[0].name, self._df.iloc[-1].name + 1)

        gids = self._df["gid"].values
        if not np.array_equal(np.sort(gids), gids):
            msg = (
                "Points are not in sequential order and will be sorted! The "
                'original order is being preserved under column "points_order"'
            )
            logger.warning(msg)
            warn(msg)
            self._df["points_order"] = self._df.index.values
            self._df = self._df.sort_values("gid")

        self._df = self._df.reset_index(drop=True)

    def __iter__(self):
        for __, row in self.df.iterrows():
            yield row

    def __getitem__(self, site_id):
        """Get the dictionary for the requested site.

        Parameters
        ----------
        site_id : int | str
            Site number (gid) of interest (typically the resource gid).

        Returns
        -------
        site : pd.Series
            Pandas Series containing information for the site with the
            requested site_id.
        """

        if site_id not in self.df["gid"].values:
            msg = (
                f"Site {site_id} not found in this instance of "
                f"ProjectPoints. Available sites include: {self.gids}"
            )
            raise gapsKeyError(msg)

        return self.df.loc[self.df["gid"] == site_id].iloc[0]

    def __repr__(self):
        return (
            f"{self.__class__.__name__} with {len(self)} sites from gid "
            f"{self.gids[0]} through {self.gids[-1]}"
        )

    def __len__(self):
        """Length of this object is the number of sites."""
        return len(self.gids)

    @property
    def df(self):
        """pd.DataFrame: Project points DataFrame of site info."""
        return self._df

    @property
    def gids(self):
        """list: Gids (resource file index values) of sites."""
        return self.df["gid"].values.tolist()

    @property
    def sites_as_slice(self):
        """list | slice: Sites in slice format or list if non-sequential."""
        # try_slice is what the sites list would be if it is sequential
        if len(self.gids) > 1:
            try_step = self.gids[1] - self.gids[0]
        else:
            try_step = 1

        try_slice = slice(self.gids[0], self.gids[-1] + 1, try_step)
        try_list = list(range(*try_slice.indices(try_slice.stop)))

        if self.gids == try_list:
            return try_slice

        # cannot be converted to a sequential slice, return list
        return self.gids

    def index(self, gid):
        """Index location (iloc not loc) for a resource gid.

        Parameters
        ----------
        gid : int
            Resource GID found in the project points gid column.

        Returns
        -------
        ind : int
            Row index of gid in the project points DataFrame.
        """
        if gid not in self._df["gid"].values:
            msg = (
                f"Requested resource gid {gid} is not present in the project "
                f"points DataFrame. Cannot return row index."
            )
            raise gapsIndexError(msg)

        ind = np.where(self._df["gid"] == gid)[0][0]

        return ind

    def join_df(self, df2, key="gid"):
        """Join df2 to the _df attribute using _df's gid as the join key.

        This can be used to add site-specific data to the project_points,
        taking advantage of the `ProjectPoints` iterator/split functions
        such that only the relevant site data is passed to the analysis
        functions.

        Parameters
        ----------
        df2 : pd.DataFrame
            DataFrame to be joined to the :attr:`df` attribute (this
            instance of project points DataFrame). This likely contains
            site-specific inputs that are to be passed to parallel
            workers.
        key : str
            Primary key of df2 to be joined to the :attr:`df` attribute
            (this instance of the project points DataFrame). Primary key
            of the self._df attribute is fixed as the gid column.
        """
        # ensure df2 doesn't have any duplicate columns for suffix reasons.
        df2_cols = [c for c in df2.columns if c not in self._df or c == key]
        self._df = pd.merge(
            self._df,
            df2[df2_cols],
            how="left",
            left_on="gid",
            right_on=key,
            copy=False,
            validate="1:1",
        )

    def get_sites_from_key(self, key, value):
        """Get a site list for which the key equals the value.

        Parameters
        ----------
        key : str
            Name of key (column) in project points DataFrame.
        value : int | float | str | obj
            Value to look for under the ``key`` column.

        Returns
        -------
        sites : lis of ints
            List of sites (GID values) associated with the requested key
            and value. If the key or value does not exist, an empty list
            is returned.
        """
        if key not in self.df:
            return []

        return list(self.df.loc[self.df[key] == value, "gid"].values)

    def split(self, sites_per_split=100):
        """Split the project points into sub-groups by number of sites.

        Parameters
        ----------
        sites_per_split : int, optional
            Number of points in each sub-group. By default, `100`.

        Yields
        ------
        ProjectPoints
            A new ProjectPoints instance with up to `sites_per_split`
            number of points.
        """
        for ind in range(0, len(self.df), sites_per_split):
            yield self.__class__(self.df.iloc[ind : ind + sites_per_split])

    @classmethod
    def from_range(cls, split_range, points, **kwargs):
        """Create a ProjectPoints instance from a range indices.

        Parameters
        ----------
        split_range : 2-tuple
            Tuple containing the start and end index (iloc, not loc).
            Last index is not included.
        points : int | slice | list | tuple | str | pd.DataFrame | dict
            Slice specifying project points, string pointing to a
            project points csv, or a DataFrame containing the effective
            csv contents. Can also be a single integer site value.
        **kwargs
            Keyword-argument pairs to add to project points DataFrame.
            The key should be the column name, and the value should
            be the value to add under the column name. Values must
            either be a scalar or match the length of the DataFrame
            resulting from the `points` input.

        Returns
        -------
        ProjectPoints
            A new ProjectPoints instance with a range sampled from the
            points input according to ``split_range``.
        """
        pp = cls(points, **kwargs)
        start, end = split_range
        return cls(pp.df.iloc[start:end])



# -*- coding: utf-8 -*-
"""GAPs batching framework for parametric runs.

Based on reV-batch.
"""
import os
import copy
import json
import shutil
import logging
from warnings import warn
from pathlib import Path
from itertools import product
from collections import namedtuple

import pandas as pd

from rex.utilities import parse_year

from gaps.config import (
    load_config,
    init_logging_from_config,
    ConfigType,
    resolve_all_paths,
)
import gaps.cli.pipeline
from gaps.pipeline import Pipeline
from gaps.exceptions import gapsValueError, gapsConfigError
from gaps.warnings import gapsWarning


logger = logging.getLogger(__name__)

BATCH_CSV_FN = "batch_jobs.csv"
BatchSet = namedtuple("BatchSet", ["arg_combo", "file_set", "tag"])


class BatchJob:
    """Framework for building a batched job suite.

    Based on reV-batch.

    This framework allows users to modify key-value pairs in input
    configuration files based on a top-level batch config file. This
    framework will create run directories for all combinations of input
    parametrics and run the corresponding GAPs pipelines for each job.
    """

    def __init__(self, config):
        """
        Parameters
        ----------
        config : str
            File path to config json or csv (str).
        """
        self._job_tags = None
        self._base_dir, config = _load_batch_config(config)
        self._pipeline_fp = Path(config["pipeline_config"])

        init_logging_from_config(config)
        self._sets = _parse_config(config)

        logger.info("Batch job initialized with %d sub jobs.", len(self._sets))

    @property
    def job_table(self):
        """pd.DataFrame: Batch job summary table."""
        jobs = []
        for job_tag, (arg_comb, file_set, set_tag) in self._sets.items():
            job_info = {k: str(v) for k, v in arg_comb.items()}
            job_info["set_tag"] = str(set_tag)
            job_info["files"] = str(file_set)
            job_info = pd.DataFrame(job_info, index=[job_tag])
            jobs.append(job_info)

        table = pd.concat(jobs)

        table.index.name = "job"
        table["pipeline_config"] = self._pipeline_fp.as_posix()

        return table

    @property
    def sub_dirs(self):
        """list: Job sub directory paths."""
        return [self._base_dir / tag for tag in self._sets]

    def _make_job_dirs(self):
        """Copy job files from the batch config dir into sub job dirs."""

        table = self.job_table
        table.to_csv(self._base_dir / BATCH_CSV_FN)
        logger.debug(
            "Batch jobs list: %s", sorted(table.index.values.tolist())
        )
        logger.info("Preparing batch job directories...")

        # walk through current directory getting everything to copy
        for source_dir, _, filenames in os.walk(self._base_dir):

            # don't make additional copies of job sub directories.
            if any(job_tag in source_dir for job_tag in self._sets):
                continue

            # For each dir level, iterate through the batch arg combos
            for tag, (arg_comb, mod_files, __) in self._sets.items():

                # Add the job tag to the directory path.
                # This will copy config subdirs into the job subdirs
                source_dir = Path(source_dir)
                index = source_dir.parts.index(self._base_dir.name)
                destination_dir = self._base_dir / tag
                destination_dir = destination_dir.joinpath(
                    *source_dir.parts[index + 1 :]
                )
                destination_dir.mkdir(parents=True, exist_ok=True)

                for name in filenames:
                    if BATCH_CSV_FN in name:
                        continue
                    _copy_batch_file(
                        source_dir / name,
                        destination_dir / name,
                    )

                for fp_source in mod_files:
                    logger.debug(
                        "Copying and modifying run file %r to job: %r",
                        fp_source,
                        tag,
                    )
                    fp_source = Path(fp_source)
                    fp_target = destination_dir / fp_source.name
                    _mod_file(fp_source, fp_target, arg_comb)

                _copy_batch_file(
                    self._pipeline_fp,
                    destination_dir / self._pipeline_fp.name,
                )

        logger.info("Batch job directories ready for execution.")

    def _run_pipelines(self, monitor_background=False):
        """Run the pipeline modules for each batch job."""

        for sub_directory in self.sub_dirs:
            pipeline_config = sub_directory / self._pipeline_fp.name
            if not pipeline_config.is_file():
                raise gapsConfigError(
                    f"Could not find pipeline config to run: "
                    f"{pipeline_config!r}"
                )
            if monitor_background:
                # pylint: disable=no-value-for-parameter
                gaps.cli.pipeline.pipeline(
                    pipeline_config,
                    cancel=False,
                    monitor=True,
                    background=True,
                )
            else:
                Pipeline.run(pipeline_config, monitor=False)

    def cancel(self):
        """Cancel all pipeline modules for all batch jobs."""
        for sub_directory in self.sub_dirs:
            pipeline_config = sub_directory / self._pipeline_fp.name
            if pipeline_config.is_file():
                Pipeline.cancel_all(pipeline_config)

    def delete(self):
        """Clear all of the batch sub job folders.

        Only the batch sub folders listed in the job summary csv file
        in the batch config directory are deleted.
        """

        fp_job_table = self._base_dir / BATCH_CSV_FN
        if not fp_job_table.exists():
            msg = (
                f"Cannot delete batch jobs without jobs summary table: "
                f"{fp_job_table!r}"
            )
            logger.error(msg)
            raise FileNotFoundError(msg)

        job_table = pd.read_csv(fp_job_table, index_col=0)

        if job_table.index.name != "job":
            msg = (
                "Cannot delete batch jobs when the batch summary table "
                'does not have "job" as the index key'
            )
            raise gapsValueError(msg)

        self._remove_sub_dirs(job_table)
        fp_job_table.unlink()

    def _remove_sub_dirs(self, job_table):
        """Remove all the sub-directories tracked in the job table."""
        for sub_dir in job_table.index:
            job_dir = self._base_dir / sub_dir
            if job_dir.exists():
                logger.info("Removing batch job directory: %r", sub_dir)
                shutil.rmtree(job_dir)
            else:
                msg = "Cannot find batch job directory: {sub_dir!r}"
                warn(msg, gapsWarning)

    def run(self, dry_run=False, monitor_background=False):
        """Run the batch job from a config file.

        Parameters
        ----------
        dry_run : bool
            Flag to make job directories without running.
        monitor_background : bool
            Flag to monitor all batch pipelines continuously
            in the background. Note that the stdout/stderr will not be
            captured, but you can set a pipeline "log_file" to capture
            logs.
        """
        self._make_job_dirs()
        if dry_run:
            return

        self._run_pipelines(monitor_background=monitor_background)


def _load_batch_config(config_fp):
    """Load and validate the batch config file (CSV or JSON)"""
    base_dir = Path(config_fp).expanduser().parent.resolve()
    config = _load_batch_config_to_dict(config_fp)
    config = _validate_batch_config(config, base_dir)
    return base_dir, config


def _load_batch_config_to_dict(config_fp):
    """Load the batch file to dict."""
    if Path(config_fp).name.endswith(".csv"):
        return _load_batch_csv(config_fp)
    return load_config(config_fp, resolve_paths=False)


def _load_batch_csv(config_fp):
    """Load batch csv file to dict."""
    table = pd.read_csv(config_fp)
    table = table.where(pd.notnull(table), None)
    _validate_batch_table(table)
    return _convert_batch_table_to_dict(table)


def _validate_batch_table(table):
    """Validate batch file CSV table."""
    if "set_tag" not in table or "files" not in table:
        msg = 'Batch CSV config must have "set_tag" and "files" columns'
        raise gapsConfigError(msg)

    set_tags_not_unique = len(table.set_tag.unique()) != len(table)
    if set_tags_not_unique:
        msg = 'Batch CSV config must have completely unique "set_tag" column'
        raise gapsConfigError(msg)

    if "pipeline_config" not in table:
        msg = (
            'Batch CSV config must have "pipeline_config" columns specifying '
            "the pipeline config filename."
        )
        raise gapsConfigError(msg)


def _convert_batch_table_to_dict(table):
    """Convert validated batch csv file to dict."""
    sets = []
    for _, job in table.iterrows():
        job_dict = job.to_dict()
        args = {
            k: [v]
            for k, v in job_dict.items()
            if k not in ("set_tag", "files", "pipeline_config")
        }
        files = json.loads(job_dict["files"].replace("'", '"'))
        set_config = {
            "args": args,
            "set_tag": job_dict["set_tag"],
            "files": files,
        }
        sets.append(set_config)

    batch_dict = {
        "logging": {"log_file": None, "log_level": "INFO"},
        "pipeline_config": table["pipeline_config"].values[0],
        "sets": sets,
    }

    return batch_dict


def _validate_batch_config(config, base_dir):
    """Validate the batch config dict."""
    config = _check_pipeline(config, base_dir)
    return _check_sets(config, base_dir)


def _check_pipeline(config, base_dir):
    """Check the pipeline config file in the batch config."""

    if "pipeline_config" not in config:
        raise gapsConfigError('Batch config needs "pipeline_config" arg!')

    config["pipeline_config"] = resolve_all_paths(
        config["pipeline_config"], base_dir
    )
    if not Path(config["pipeline_config"]).exists():
        raise gapsConfigError(
            f"Could not find the pipeline config file: "
            f"{config['pipeline_config']!r}"
        )
    return config


def _check_sets(config, base_dir):
    """Check the batch sets for required inputs and valid files."""

    if "sets" not in config:
        raise gapsConfigError('Batch config needs "sets" arg!')

    batch_sets = []
    for batch_set in config["sets"]:
        if not isinstance(batch_set, dict):
            raise gapsConfigError("Batch sets must be dictionaries.")
        if "args" not in batch_set:
            raise gapsConfigError('All batch sets must have "args" key.')
        if "files" not in batch_set:
            raise gapsConfigError('All batch sets must have "files" key.')
        batch_set["files"] = resolve_all_paths(batch_set["files"], base_dir)
        for fpath in batch_set["files"]:
            if not Path(fpath).exists():
                raise gapsConfigError(
                    "Could not find file to modify in batch jobs: {fpath!r}"
                )
        batch_sets.append(batch_set)

    config["sets"] = batch_sets
    return config


def _enumerated_product(args):
    """An enumerated product function."""
    yield from zip(product(*(range(len(x)) for x in args)), product(*args))


def _parse_config(config):
    """Parse batch config object for useful data."""

    sets = set()
    batch_sets = {}

    for batch_set in config["sets"]:
        set_tag = batch_set.get("set_tag", "")
        args = batch_set["args"]
        if set_tag in sets:
            msg = f"Found multiple sets with the same set_tag: {set_tag!r}"
            raise gapsValueError(msg)
        sets.add(set_tag)

        for inds, comb in _enumerated_product(args.values()):
            arg_combo = dict(zip(args, comb))
            arg_inds = dict(zip(args, inds))
            tag_arg_comb = {
                k: v for k, v in arg_combo.items() if len(args[k]) > 1
            }
            job_tag = _make_job_tag(set_tag, tag_arg_comb, arg_inds)
            batch_sets[job_tag] = BatchSet(
                arg_combo, batch_set["files"], set_tag
            )

    return batch_sets


def _make_job_tag(set_tag, arg_comb, arg_inds):
    """Make a job tags from a unique combination of args + values."""

    job_tag = [set_tag] if set_tag else []

    for arg, value in arg_comb.items():
        arg_tag = "".join(
            [s[0] for s in arg.split("_")]
            + (
                [_format_value(value)]
                if isinstance(value, (int, float))
                else [str(arg_inds[arg])]
            )
        )

        job_tag.append(arg_tag)

    return "_".join(job_tag).rstrip("_")


def _format_value(value):
    """Format the input value as a string."""

    value = str(value).replace(".", "")

    if parse_year(f"_{value}", option="bool"):
        value = f"{value}0"

    return value


def _mod_file(fpath_in, fpath_out, arg_mods):
    """Import and modify the contents of a json. Dump to new file."""
    config_type = ConfigType(fpath_in.name.split(".")[-1])
    config = config_type.load(fpath_in)
    config_type.write(fpath_out, _mod_dict(config, arg_mods))


def _mod_dict(inp, arg_mods):
    """Recursively modify key/value pairs in a dictionary."""

    out = copy.deepcopy(inp)

    if isinstance(inp, dict):
        for key, val in inp.items():
            if key in arg_mods:
                out[key] = _clean_arg(arg_mods[key])
            elif isinstance(val, (list, dict)):
                out[key] = _mod_dict(val, arg_mods)

    elif isinstance(inp, list):
        return [_mod_dict(entry, arg_mods) for entry in inp]

    return out


def _clean_arg(arg):
    """Perform any cleaning steps required before writing an arg to a json."""

    if not isinstance(arg, str):
        return arg

    missing_bracket_pair = "{" not in arg or "}" not in arg
    missing_string_quotes = "'" not in arg and '"' not in arg
    if missing_bracket_pair or missing_string_quotes:
        return arg

    arg = arg.replace("'", '"')
    try:
        return json.loads(arg)
    except json.decoder.JSONDecodeError as err:
        logger.exception("Could not load json string: %s", arg)
        raise err


def _copy_batch_file(fp_source, fp_target):
    """Copy a file in the batch directory into a job directory if needed."""
    if not _source_needs_copying(fp_source, fp_target):
        return

    logger.debug("Copying run file %r to %r", fp_source, fp_target)
    shutil.copy(fp_source, fp_target)


def _source_needs_copying(fp_source, fp_target):
    """Determine if source needs to be copied to dest."""
    if not fp_target.exists():
        return True
    return fp_source.lstat().st_mtime > fp_target.lstat().st_mtime



# -*- coding: utf-8 -*-
"""
GAPs config preprocessing functions.
"""
import re
import glob
from math import ceil
from warnings import warn
from collections import abc

from gaps.project_points import ProjectPoints
from gaps.cli.config import TAG
from gaps.pipeline import parse_previous_status
from gaps.utilities import resolve_path
from gaps.exceptions import gapsConfigError

from gaps.warnings import gapsWarning


def split_project_points_into_ranges(config):
    """Split project points into ranges inside of config.

    Parameters
    ----------
    config : dict
        Run config. This config must have a "project_points" input that
        can be used to initialize :class:`ProjectPoints`.

    Returns
    -------
    dict
        Run config with a "project_points_split_range" key that split
        the project points into multiple ranges based on node input.
    """
    project_points_file = config["project_points"]
    exec_control = config.get("execution_control", {})
    if exec_control.get("option") == "local":
        num_nodes = 1
    else:
        num_nodes = exec_control.pop("nodes", 1)

    points = ProjectPoints(project_points_file)
    sites_per_split = ceil(len(points) / num_nodes)
    config["project_points_split_range"] = [
        sub_points.split_range
        for sub_points in points.split(sites_per_split=sites_per_split)
    ]
    return config


def preprocess_collect_config(
    config, project_dir, command_name, collect_pattern="PIPELINE"
):
    """Pre-process collection config.

    Specifically, the "collect_pattern" key is resolved into a list of
    2-tuples, where the first element in each tuple is a path to the
    collection output file, while the second element is the
    corresponding filepath-pattern representing the files to be
    collected into the output file.

    Parameters
    ----------
    config : dict
        Collection config. This config will be updated to include a
        "collect_pattern" key if it doesn't already have one. If the
        "collect_pattern" key exists, it can be a string, a collection,
        or a mapping with a `.items()` method.
    project_dir : path-like
        Path to project directory. This path is used to resolve the
        out filepath input from the user.
    command_name : str
        Name of the command being run. This is used to parse the
        pipeline status for output files if "collect_pattern"="PIPELINE"
        in the input `config`.
    collect_pattern : str | list | dict, optional
        Unix-style /filepath/pattern*.h5 representing the files to be
        collected into a single output HDF5 file. If no output file path
        is specified (i.e. this input is a single pattern or a list of
        patterns), the output file path will be inferred from the
        pattern itself (specifically, the wildcard will be removed
        and the result will be the output file path). If a list of
        patterns is provided, each pattern will be collected into a
        separate output file. To specify the name of the output file(s),
        set this input to a dictionary where the keys are paths to the
        output file (including the filename itself; relative paths are
        allowed) and the values are patterns representing the files that
        should be collected into the output file. If running a collect
        job as part of a pipeline, this input can be set to "PIPELINE",
        which will parse the output of the previous step and generate
        the input file pattern and output file name automatically.
        By default, ``"PIPELINE"``.

    Returns
    -------
    dict
        Updated collection config.
    """
    files = collect_pattern
    if files == "PIPELINE":
        files = parse_previous_status(project_dir, command_name)
        files = [re.sub(f"{TAG}\\d+", "*", fname) for fname in files]

    if isinstance(files, str):
        files = [files]

    if isinstance(files, abc.Sequence):
        files = {pattern.replace("*", ""): pattern for pattern in files}

    files = [
        (
            resolve_path(
                out_path if out_path.startswith("/") else f"./{out_path}",
                project_dir,
            ),
            pattern,
        )
        for out_path, pattern in files.items()
    ]
    config["_out_path"], config["_pattern"] = zip(*_validate_patterns(files))
    return config


def _validate_patterns(files):
    """Remove any patterns that have no corresponding files."""

    patterns_with_no_files = []
    files_to_collect = []
    for out, pattern in files:
        if glob.glob(pattern):
            files_to_collect.append((out, pattern))
        else:
            patterns_with_no_files.append(pattern)

    if any(patterns_with_no_files):
        msg = (
            f"Could not find any files for the following patterns: "
            f"{patterns_with_no_files}. Skipping..."
        )
        warn(msg, gapsWarning)

    if not files_to_collect:
        msg = (
            "Found no files to collect! Please double check your config input "
            "and verify that the files to be collected exist on disk!"
        )
        raise gapsConfigError(msg)

    return files_to_collect



# -*- coding: utf-8 -*-
"""GAPs Status Monitor"""
from pathlib import Path
from warnings import warn
from itertools import chain

import click
import psutil
from colorama import init, Fore, Style
from tabulate import tabulate

from gaps.status import Status, StatusField, StatusOption, _elapsed_time_as_str
from gaps.warnings import gapsWarning
from gaps.cli.command import _WrappedCommand


FAILURE_STRINGS = ["failure", "fail", "failed", "f"]
RUNNING_STRINGS = ["running", "run", "r"]
SUBMITTED_STRINGS = ["submitted", "submit", "sb", "pending", "pend", "p"]
SUCCESS_STRINGS = ["successful", "success", "s"]
# cspell:disable-next-line
NOT_SUBMITTED_STRINGS = ["unsubmitted", "unsubmit", "u", "not_submitted", "ns"]
FILTER_HELP = """
Filter jobs for the requested status(es). Allowed options (case-insensitive)
include:

    - Failed: {fail_options}
    - Running: {running_options}
    - Submitted: {submitted_options}
    - Success: {success_options}
    - Not submitted: {ns_options}

Multiple status keys can be specified by repeating this option
(e.g. :code:`-s status1 -s status2 ...`). By default, all status values are
displayed.
"""
STATUS_HELP = """
Display the status of a project FOLDER.

By default, the status of the current working directory is displayed.
"""


def _filter_df_for_status(df, status_request):
    """Check for a specific status."""

    filter_statuses = set()
    for request in status_request:
        request = request.lower()
        if request in FAILURE_STRINGS:
            filter_statuses |= {StatusOption.FAILED}
        elif request in SUCCESS_STRINGS:
            filter_statuses |= {StatusOption.SUCCESSFUL}
        elif request in RUNNING_STRINGS:
            filter_statuses |= {StatusOption.RUNNING}
        elif request in SUBMITTED_STRINGS:
            filter_statuses |= {StatusOption.SUBMITTED}
        elif request in NOT_SUBMITTED_STRINGS:
            filter_statuses |= {StatusOption.NOT_SUBMITTED}
        else:
            msg = (
                f"Requested status not recognized: {status_request!r}. "
                "No additional filtering performed!"
            )
            warn(msg, gapsWarning)

    df = df[df[StatusField.JOB_STATUS].isin(filter_statuses)]
    return df.copy()


def _color_print(df, print_folder, commands, status):
    """Color the status portion of the print out."""

    def color_string(string):
        if string == StatusOption.FAILED:
            string = f"{Fore.RED}{string}{Style.RESET_ALL}"
        elif string == StatusOption.SUCCESSFUL:
            string = f"{Fore.GREEN}{string}{Style.RESET_ALL}"
        elif string == StatusOption.RUNNING:
            string = f"{Fore.BLUE}{string}{Style.RESET_ALL}"
        else:
            string = f"{Fore.YELLOW}{string}{Style.RESET_ALL}"
        return string

    extras = []
    commands = ", ".join(commands or [])
    if commands:
        extras.append(f"commands={commands}")
    status = ", ".join(status or [])
    if status:
        extras.append(f"status={status}")
    if extras:
        extras = "; ".join(extras)
        extras = f" ({extras})"
    else:
        extras = ""

    pid = df.monitor_pid
    total_runtime_seconds = df.total_runtime_seconds
    name = f"\n{Fore.CYAN}{print_folder}{extras}{Style.RESET_ALL}:"
    job_status = StatusField.JOB_STATUS
    df[job_status.value] = df[job_status].apply(color_string)
    df = df.fillna("--")
    pdf = tabulate(
        df,
        showindex=True,  # cspell:disable-line
        headers=df.columns,
        tablefmt="simple",  # cspell:disable-line
        disable_numparse=[3],  # cspell:disable-line
    )
    print(name)
    if pid is not None and psutil.pid_exists(pid):
        print(f"{Fore.MAGENTA}MONITOR PID: {pid}{Style.RESET_ALL}")
    print(pdf)
    if total_runtime_seconds is not None:
        runtime_str = (
            f"Total Runtime: {_elapsed_time_as_str(total_runtime_seconds)}"
        )
        # TODO: make prettier
        # divider = "".join(["-"] * len(runtime_str))
        # divider = "".join(["~"] * 3)
        divider = ""
        print(divider)
        print(f"{Style.BRIGHT}{runtime_str}{Style.RESET_ALL}")
        print(
            f"{Style.BRIGHT}**Statistics only include shown jobs, excluding "
            f"failed or duplicate runs**{Style.RESET_ALL}"
        )
    print()


def main_monitor(folder, commands, status, include):
    """Run the appropriate monitor functions for a folder.

    Parameters
    ----------
    folder : path-like
        Path to folder for which to print status.
    commands : container of str
        Container with the commands to display.
    status : container of str
        Container with the statuses to display.
    include : container of str
        Container of extra keys to pull from the status files.
    """

    init()
    folder = Path(folder).expanduser().resolve()
    for directory in chain([folder], folder.rglob("*")):
        if not directory.is_dir():
            continue

        pipe_status = Status(directory)
        if not pipe_status:
            continue

        include_with_runtime = list(include) + [StatusField.RUNTIME_SECONDS]
        df = pipe_status.as_df(
            commands=commands, include_cols=include_with_runtime
        )
        if status:
            df = _filter_df_for_status(df, status)

        total_runtime = df[StatusField.RUNTIME_SECONDS].sum()
        df = df[list(df.columns)[:-1]]
        df.monitor_pid = pipe_status.get(StatusField.MONITOR_PID)
        df.total_runtime_seconds = total_runtime
        _color_print(df, directory.name, commands, status)


def status_command():
    """A status CLI command."""
    filter_help = FILTER_HELP.format(
        fail_options=" ".join([f"``{s}``" for s in FAILURE_STRINGS]),
        running_options=" ".join([f"``{s}``" for s in RUNNING_STRINGS]),
        submitted_options=" ".join([f"``{s}``" for s in SUBMITTED_STRINGS]),
        success_options=" ".join([f"``{s}``" for s in SUCCESS_STRINGS]),
        ns_options=" ".join([f"``{s}``" for s in NOT_SUBMITTED_STRINGS]),
    )

    params = [
        click.Argument(
            param_decls=["folder"],
            default=Path.cwd(),
            type=click.Path(exists=True),
        ),
        click.Option(
            param_decls=["--commands", "-c"],
            multiple=True,
            default=None,
            help="Filter status for the given command(s). Multiple commands "
            "can be specified by repeating this option (e.g. :code:`-c "
            "command1 -c command2 ...`) By default, the status of all "
            "commands is displayed.",
        ),
        click.Option(
            param_decls=["--status", "-s"],
            multiple=True,
            default=None,
            help=filter_help,
        ),
        click.Option(
            param_decls=["--include", "-i"],
            multiple=True,
            default=None,
            help="Extra status keys to include in the print output for each "
            "job. Multiple status keys can be specified by repeating "
            "this option (e.g. :code:`-i key1 -i key2 ...`) By default, no "
            "extra keys are displayed.",
        ),
    ]

    return _WrappedCommand(
        "status",
        context_settings=None,
        callback=main_monitor,
        params=params,
        help=STATUS_HELP,
        epilog=None,
        short_help=None,
        options_metavar="[OPTIONS]",
        add_help_option=True,
        no_args_is_help=False,
        hidden=False,
        deprecated=False,
    )



"""
setup.py
"""
from pathlib import Path

from setuptools import setup, find_packages


HERE = Path(__file__).parent.resolve()

with open(HERE / "README.rst", encoding="utf-8") as f:
    README = f.read()

with open(HERE / "gaps" / "version.py", encoding="utf-8") as f:
    VERSION = f.read().split("=")[-1].strip().strip('"').strip("'")

with open("requirements.txt") as f:
    INSTALL_REQUIREMENTS = f.read().splitlines()


DEV_REQUIREMENTS = ["black", "pylint", "jupyter", "pipreqs"]
TEST_REQUIREMENTS = ["pytest", "pytest-cov", "h5py"]
DOC_REQUIREMENTS = [
    "make",
    "ghp-import",
    "numpydoc",
    "pandoc",
    "sphinx_rtd_theme>=0.5.1",
]
DESCRIPTION = (
    "National Renewable Energy Laboratory's (NREL's) Geospatial Analysis "
    "Pipelines (GAPs) framework"
)


setup(
    name="NREL-gaps",
    version=VERSION,
    description=DESCRIPTION,
    long_description=README,
    long_description_content_type="text/markdown",
    author="Paul Pinchuk",
    maintainer_email="ppinchuk@nrel.gov",
    url="https://nrel.github.io/gaps/",
    packages=find_packages(),
    package_dir={"gaps": "gaps"},
    license="BSD 3-Clause",
    zip_safe=False,
    keywords="gaps",
    classifiers=[
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.10",
    ],
    test_suite="tests",
    install_requires=INSTALL_REQUIREMENTS,
    extras_require={
        "test": TEST_REQUIREMENTS,
        "dev": TEST_REQUIREMENTS + DEV_REQUIREMENTS,
        "docs": TEST_REQUIREMENTS + DEV_REQUIREMENTS + DOC_REQUIREMENTS,
    },
)



# -*- coding: utf-8 -*-
# pylint: disable=invalid-name,arguments-differ,too-many-arguments
"""
HPC Execution utilities.
"""
import re
import shlex
import logging
import getpass
import subprocess
from math import floor
from pathlib import Path
from warnings import warn
from collections import namedtuple
from abc import ABC, abstractmethod


from gaps.exceptions import gapsExecutionError, gapsHPCError, gapsValueError
from gaps.warnings import gapsHPCWarning


logger = logging.getLogger(__name__)


Q_COLUMNS = namedtuple("Q_COLUMNS", ["NAME", "ID", "STATUS"])
COMMANDS = namedtuple("COMMANDS", ["SUBMIT", "CANCEL"])
DEFAULT_STDOUT_PATH = "./stdout"
"""Default directory for .stdout and .stderr files."""


class HpcJobManager(ABC):
    """Abstract HPC job manager framework"""

    # get username as class attribute.
    USER = getpass.getuser()

    # set a max job name length, will raise error if too long.
    MAX_NAME_LEN = 100

    SHELL_FILENAME_FMT = "{}.sh"

    def __init__(self, user=None, queue_dict=None):
        """
        Parameters
        ----------
        user : str | None, optional
            HPC username. `None` will get your username using
            :func:`getpass.getuser`. By default, `None`.
        queue_dict : dict | None, optional
            Parsed HPC queue dictionary from :meth:`parse_queue_str`.
            `None` will get the queue info from the hardware.
            By default, `None`.
        """

        self._user = user or self.USER
        if queue_dict is not None and not isinstance(queue_dict, dict):
            msg = (
                f"HPC queue_dict arg must be None or Dict but received: "
                f"{queue_dict}, {type(queue_dict)}"
            )
            raise gapsHPCError(msg)

        self._queue = queue_dict

    @classmethod
    def parse_queue_str(cls, queue_str):
        """Parse the hardware queue string into a nested dictionary.

        This function parses the queue output string into a dictionary
        keyed by integer job ids with values as dictionaries of job
        properties (queue printout columns).

        Parameters
        ----------
        queue_str : str
            HPC queue output string. Typically a space-delimited string
            with line breaks.

        Returns
        -------
        queue_dict : dict
            HPC queue parsed into dictionary format keyed by integer job
            ids with values as dictionaries of job properties (queue
            printout columns).
        """

        queue_dict = {}
        header, queue_rows = cls._split_queue(queue_str)
        for row in queue_rows:
            job = [k.strip(" ") for k in row.strip(" ").split(" ") if k != ""]
            job_id = int(job[header.index(cls.COLUMN_HEADERS.ID)])
            queue_dict[job_id] = dict(zip(header, job))

        return queue_dict

    @property
    def queue(self):
        """dict: HPC queue keyed by job ids with values as job properties."""
        if self._queue is None:
            queue_str = self.query_queue()
            self._queue = self.parse_queue_str(queue_str)

        return self._queue

    def check_status_using_job_id(self, job_id):
        """Check the status of a job using the HPC queue and job ID.

        Parameters
        ----------
        job_id : int
            Job integer ID number.

        Returns
        -------
        status : str | None
            Queue job status string or `None` if not found.
        """
        job_id = int(job_id)
        return self.queue.get(job_id, {}).get(self.COLUMN_HEADERS.STATUS)

    def check_status_using_job_name(self, job_name):
        """Check the status of a job using the HPC queue and job name.

        Parameters
        ----------
        job_name : str
            Job name string.

        Returns
        -------
        status : str | None
            Queue job status string or `None` if not found.
        """
        for attrs in self.queue.values():
            if attrs[self.COLUMN_HEADERS.NAME] == job_name:
                return attrs[self.COLUMN_HEADERS.STATUS]

        return None

    def cancel(self, arg):
        """Cancel a job.

        Parameters
        ----------
        arg : int | list | str
            Integer job id(s) to cancel. Can be a list of integer
            job ids, 'all' to cancel all jobs, or a feature (-p short)
            to cancel all jobs with a given feature
        """
        self._validate_command_not_none("CANCEL")

        if isinstance(arg, (list, tuple)):
            for job_id in arg:
                self.cancel(job_id)

        elif str(arg).lower() == "all":
            self._queue = None
            for job_id in self.queue.keys():
                self.cancel(job_id)

        elif isinstance(arg, (int, str)):
            cmd = f"{self.COMMANDS.CANCEL} {arg}"
            cmd = shlex.split(cmd)
            subprocess.call(cmd)

        else:
            raise gapsExecutionError(
                f"Could not cancel: {arg} with type {type(arg)}"
            )

    def submit(
        self,
        name,
        keep_sh=False,
        **kwargs,
    ):
        """Submit a job on the HPC.

        Parameters
        ----------
        name : str
            HPC job name.
        keep_sh : bool, optional
            Option to keep the submission script on disk.
            By default, `False`.
        **kwargs
            Extra keyword-argument pairs to be passed to
            :meth:`make_script_str`.

        Returns
        -------
        out : str
            Standard output from submission. If submitted successfully,
            this is the Job ID.
        err : str
            Standard error. This is an empty string if the job was
            submitted successfully.
        """
        self._validate_command_not_none("SUBMIT")
        self._validate_name_length(name)
        if self._job_is_running(name):
            logger.info(
                "Not submitting job %r because it is already in "
                "queue has been recently submitted",
                name,
            )
            return None, "already_running"

        self._setup_submission(name, **kwargs)

        out, err = submit(
            f"{self.COMMANDS.SUBMIT} {self.SHELL_FILENAME_FMT.format(name)}"
        )
        out = self._teardown_submission(name, out, err, keep_sh=keep_sh)
        return out, err

    def _validate_command_not_none(self, command):
        """Validate that a command is not `None`."""
        if getattr(self.COMMANDS, command) is None:
            raise NotImplementedError(
                f"{command!r} command has not been defined for class"
                f"{self.__class__.__name__!r}"
            )

    def _validate_name_length(self, name):
        """Validate that the name does not exceed max length."""
        if len(name) > self.MAX_NAME_LEN:
            raise gapsValueError(
                f"Cannot submit job with name longer than "
                f"{self.MAX_NAME_LEN} chars: {name!r}"
            )

    def _setup_submission(self, name, **kwargs):
        """Setup submission file and directories."""
        stdout_dir = Path(kwargs.get("stdout_path", DEFAULT_STDOUT_PATH))
        stdout_dir.mkdir(parents=True, exist_ok=True)

        script = self.make_script_str(name, **kwargs)

        make_sh(self.SHELL_FILENAME_FMT.format(name), script)

    def _teardown_submission(self, name, out, err, keep_sh=False):
        """Remove submission file and mark job as submitted."""
        if not keep_sh:
            Path(self.SHELL_FILENAME_FMT.format(name)).unlink()

        if err:
            warn(
                f"Received an error or warning during submission: {err}",
                gapsHPCWarning,
            )
        else:
            out = self._mark_job_as_submitted(name, out)

        return out

    def _mark_job_as_submitted(self, name, out):
        """Mark job as submitted in the queue."""
        job_id = int(out.split(" ")[-1])
        out = str(job_id)
        logger.debug("Job %r with id #%s submitted successfully", name, job_id)
        self._queue[job_id] = {
            self.COLUMN_HEADERS.ID: job_id,
            self.COLUMN_HEADERS.NAME: name,
            self.COLUMN_HEADERS.STATUS: self.Q_SUBMITTED_STATUS,
        }
        return out

    @abstractmethod
    def _job_is_running(self, name):
        """Check wether the job is submitted/running."""
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def _split_queue(cls, queue_str):
        """Split queue string into a header and a list of lines."""
        raise NotImplementedError

    @abstractmethod
    def query_queue(self):
        """Run the HPC queue command and return the raw stdout string.

        Returns
        -------
        stdout : list
            HPC queue output string that can be split into a list on
            line breaks.
        """
        raise NotImplementedError

    @abstractmethod
    def make_script_str(self, name, **kwargs):
        """Generate the submission script."""
        raise NotImplementedError

    @property
    @abstractmethod
    def COLUMN_HEADERS(self):
        """`namedtuple`: Column header names."""
        raise NotImplementedError

    @property
    @abstractmethod
    def COMMANDS(self):
        """`namedtuple`: Command names."""
        raise NotImplementedError

    @property
    @abstractmethod
    def Q_SUBMITTED_STATUS(self):
        """str: String representing the submitted status for this manager."""
        raise NotImplementedError


class PBS(HpcJobManager):
    """Subclass for PBS subprocess jobs."""

    COLUMN_HEADERS = Q_COLUMNS(NAME="Name", ID="Job id", STATUS="S")

    # String representing the submitted status for this manager
    Q_SUBMITTED_STATUS = "Q"

    COMMANDS = COMMANDS(SUBMIT="qsub", CANCEL="qdel")  # cspell:disable-line

    def query_queue(self):
        """Run the PBS qstat command and return the raw stdout string.

        Returns
        -------
        stdout : str
            qstat output string. Can be split on line breaks to get
            list.
        """
        stdout, _ = submit(f"qstat -u {self._user}")
        return _skip_q_rows(stdout, skip_rows=(0, 1))

    @classmethod
    def _split_queue(cls, queue_str):
        """Split queue into a header and a list of lines."""
        header = ("Job id", "Name", "User", "Time Use", "S", "Queue")
        return header, queue_str.split("\n")

    def _job_is_running(self, name):
        """Check wether the job is submitted/running."""
        return self.check_status_using_job_name(name) in {"Q", "R"}

    def make_script_str(
        self,
        name,
        cmd,
        allocation,
        walltime,
        memory=None,
        queue=None,
        feature=None,
        stdout_path=DEFAULT_STDOUT_PATH,
        conda_env=None,
        sh_script=None,
    ):
        """Generate the PBS submission script.

        Parameters
        ----------
        name : str
            PBS job name.
        cmd :  str
            Command to be submitted in PBS shell script. Example:
                'python -m reV.generation.cli_gen'
        allocation : str
            HPC allocation account. Example: 'rev'.
        walltime : int | float
            Node walltime request in hours. Example: 4.
        memory : int , optional
            Node memory request in GB. By default, `None`.
        queue : str
            HPC queue to submit job to. Examples include: 'debug',
            'short', 'batch', 'batch-h', 'long', etc. By default,
            `None`, which uses `test_queue`.
        feature : str, optional
            PBS feature request (-l {feature}). Example:
            'feature=24core', 'qos=high', etc... By default, `None`.
        stdout_path : str, optional
            Path to print .stdout and .stderr files.
            By default, :attr:`DEFAULT_STDOUT_PATH`.
        conda_env : str, optional
            Conda environment to activate. By default, `None`.
        sh_script : str, optional
            Script to run before executing command. By default, `None`.

        Returns
        -------
        str
            PBS script to submit.
        """
        features = [
            str(feature).replace(" ", "") if feature else "",
            f"walltime={format_walltime(walltime)}" if walltime else "",
            f"mem={memory}gb" if memory else "",
        ]
        features = ",".join(filter(None, features))
        script_args = [
            "#!/bin/bash",
            f"#PBS -N {name} # job name",
            f"#PBS -A {allocation} # allocation account",
            f"#PBS -q {queue} # queue" if queue else "",
            # cspell:disable-next-line
            f"#PBS -o {stdout_path}/{name}_$PBS_JOBID.o",
            # cspell:disable-next-line
            f"#PBS -e {stdout_path}/{name}_$PBS_JOBID.e",
            f"#PBS -l {features}" if features else "",
            format_env(conda_env),
            # cspell:disable-next-line
            "echo Running on: $HOSTNAME, Machine Type: $MACHTYPE",
            "echo Running python in directory `which python`",
            sh_script,
            cmd,
        ]
        return "\n".join(filter(None, script_args))


class SLURM(HpcJobManager):
    """Subclass for SLURM subprocess jobs."""

    # cspell:disable-next-line
    COLUMN_HEADERS = Q_COLUMNS(NAME="NAME", ID="JOBID", STATUS="ST")

    # String representing the submitted status for this manager
    Q_SUBMITTED_STATUS = "PD"

    # cspell:disable-next-line
    COMMANDS = COMMANDS(SUBMIT="sbatch", CANCEL="scancel")

    def query_queue(self):
        """Run the HPC queue command and return the raw stdout string.

        Returns
        -------
        stdout : str
            HPC queue output string. Can be split on line breaks to get
            a list.
        """
        cmd = (
            f'squeue -u {self._user} --format="%.15i %.30P '
            f'%.{self.MAX_NAME_LEN}j %.20u %.10t %.15M %.25R %q"'
        )
        stdout, _ = submit(cmd)
        return _skip_q_rows(stdout)

    @classmethod
    def _split_queue(cls, queue_str):
        """Split queue into a header and a list of lines."""
        queue_rows = queue_str.split("\n")
        header = [
            k.strip(" ")
            for k in queue_rows[0].strip(" ").split(" ")
            if k != ""
        ]
        return header, queue_rows[1:]

    def _job_is_running(self, name):
        """Check wether the job is submitted/running."""
        return self.check_status_using_job_name(name) is not None

    def make_script_str(
        self,
        name,
        cmd,
        allocation,
        walltime,
        memory=None,
        feature=None,
        stdout_path=DEFAULT_STDOUT_PATH,
        conda_env=None,
        sh_script=None,
    ):
        """Generate the SLURM submission script.

        Parameters
        ----------
        name : str
            SLURM job name.
        cmd : str
            Command to be submitted in SLURM shell script. Example:
                'python -m reV.generation.cli_gen'
        allocation : str
            HPC allocation account. Example: 'rev'.
        walltime : int | float
            Node walltime request in hours. Example: 4.
        memory : int , optional
            Node memory request in GB. By default, `None`.
        feature : str, optional
            Additional flags for SLURM job. Format is "--qos=high"
            or "--depend=[state:job_id]". By default, `None`.
        stdout_path : str, optional
            Path to print .stdout and .stderr files.
            By default, :attr:`DEFAULT_STDOUT_PATH`.
        conda_env : str, optional
            Conda environment to activate. By default, `None`.
        sh_script : str, optional
            Script to run before executing command. By default, `None`.

        Returns
        -------
        str
            SLURM script to submit.
        """
        walltime = format_walltime(walltime)
        memory = memory * 1000 if memory else None

        script_args = [
            "#!/bin/bash",
            f"#SBATCH --account={allocation}" if allocation else "",
            f"#SBATCH --time={walltime}" if walltime else "",
            f"#SBATCH --job-name={name}  # job name",
            "#SBATCH --nodes=1  # number of nodes",
            f"#SBATCH --output={stdout_path}/{name}_%j.o",
            f"#SBATCH --error={stdout_path}/{name}_%j.e",
            f"#SBATCH {feature}  # extra feature" if feature else "",
            f"#SBATCH --mem={memory}  # node RAM in MB" if memory else "",
            format_env(conda_env),
            # cspell:disable-next-line
            "echo Running on: $HOSTNAME, Machine Type: $MACHTYPE",
            "echo Running python in directory `which python`",
            sh_script,
            cmd,
        ]
        return "\n".join(filter(None, script_args))


def make_sh(fname, script):
    """Make a shell script (.sh file) to execute a subprocess.

    Parameters
    ----------
    fname : str
        Name of the .sh file to create.
    script : str
        Contents to be written into the .sh file.
    """
    logger.debug(
        "The shell script %(n)r contains the following:\n"
        "~~~~~~~~~~ %(n)s ~~~~~~~~~~\n"
        "%(s)s\n"
        "~~~~~~~~~~ %(n)s ~~~~~~~~~~",
        {"n": fname, "s": script},
    )
    with open(fname, "w+") as f:
        f.write(script)


def _subprocess_popen(cmd):
    """Open a subprocess popen constructor and submit a command.

    Parameters
    ----------
    cmd : str
        Command to be submitted using python subprocess.

    Returns
    -------
    stdout : str
        Subprocess standard output. This is decoded from the subprocess
        stdout with rstrip.
    stderr : str
        Subprocess standard error. This is decoded from the subprocess
        stderr with rstrip. After decoding/rstrip, this will be empty if
        the subprocess doesn't return an error.
    """

    cmd = shlex.split(cmd)

    # pylint: disable=consider-using-with
    # use subprocess to submit command and get piped o/e
    process = subprocess.Popen(
        cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    stdout, stderr = process.communicate()
    stderr = stderr.decode("ascii").rstrip()
    stdout = stdout.decode("ascii").rstrip()

    if process.returncode != 0:
        raise OSError(
            f"Subprocess submission failed with return code "
            f"{process.returncode} and stderr:\n{stderr}"
        )

    return stdout, stderr


def _subprocess_run(cmd, background_stdout=False):
    """Open a subprocess and submit a command.

    Parameters
    ----------
    cmd : str
        Command to be submitted using python subprocess.
    background_stdout : bool
        Flag to capture the stdout/stderr from the background process
        in a nohup.out file.
    """

    nohup_cmd_fmt = ["nohup {}"]
    if not background_stdout:
        nohup_cmd_fmt += ["</dev/null >/dev/null 2>&1"]
    nohup_cmd_fmt += ["&"]

    cmd = " ".join(nohup_cmd_fmt).format(cmd)
    subprocess.run(cmd, shell=True, check=True)


def submit(cmd, background=False, background_stdout=False):
    """Open a subprocess and submit a command.

    Parameters
    ----------
    cmd : str
        Command to be submitted using python subprocess.
    background : bool
        Flag to submit subprocess in the background. stdout stderr will
        be empty strings if this is True.
    background_stdout : bool
        Flag to capture the stdout/stderr from the background process
        in a nohup.out file.

    Returns
    -------
    stdout : str
        Subprocess standard output. This is decoded from the subprocess
        stdout with rstrip.
    stderr : str
        Subprocess standard error. This is decoded from the subprocess
        stderr with rstrip. After decoding/rstrip, this will be empty if
        the subprocess doesn't return an error.
    """

    if background:
        _subprocess_run(cmd, background_stdout=background_stdout)
        return "", ""

    stdout, stderr = _subprocess_popen(cmd)
    return _job_id_or_out(stdout), stderr


def format_walltime(hours=None):
    """Get the SLURM walltime string in format "HH:MM:SS"

    Parameters
    ----------
    hours : float | int, optional
        Requested number of job hours. By default, `None`, which returns
        an empty string.

    Returns
    -------
    walltime : str
        SLURM walltime request in format "#SBATCH --time=HH:MM:SS"
    """
    if hours is not None:
        m_str = f"{round(60 * (hours % 1)):02d}"
        h_str = f"{floor(hours):02d}"
        return f"{h_str}:{m_str}:00"

    return hours


def format_env(conda_env=None):
    """Get special sbatch request strings for SLURM conda environments.

    Parameters
    ----------
    conda_env : str, optional
        Conda environment to activate. By default, `None`, which returns
        an empty string.

    Returns
    -------
    env_str : str
        SBATCH shell script source activate environment request string.
    """

    if conda_env is not None:
        return (
            f"echo source activate {conda_env}\n"
            f"source activate {conda_env}\n"
            f"echo conda env activate complete!\n"
        )

    return ""


def _skip_q_rows(queue_str, skip_rows=None):
    """Remove rows from the queue_str that are to be skipped.

    Parameters
    ----------
    queue_str : str
        HPC queue output string. Can be split on line breaks to get a
        list.
    skip_rows : list | None, optional
        Optional row index values to skip.

    Returns
    -------
    queue_str : str
        HPC queue output string. Can be split on line breaks to get a
        list.
    """
    if skip_rows is None:
        return queue_str

    queue_str = [
        row
        for i, row in enumerate(queue_str.split("\n"))
        if i not in skip_rows
    ]

    return "\n".join(queue_str)


def _job_id_or_out(input_str):
    """Extract job id from input.

    Specifically, this function checks the input for a job id and
    returns just the job id if present. Otherwise, the full input is
    returned.

    Parameters
    ----------
    input_str : str
        Stdout from :func:`submit`

    Returns
    -------
    str
        If job id is present in `input_str` then this is just the job
        id. Otherwise, the full `input_str` is returned.
    """
    return re.sub("[^0-9]", "", str(input_str)) or input_str



# -*- coding: utf-8 -*-
# pylint: disable=invalid-name
"""Custom Exceptions and Errors for gaps. """
import logging

logger = logging.getLogger("gaps")


class gapsError(Exception):
    """Generic gaps Error."""

    def __init__(self, *args, **kwargs):
        """Init exception and broadcast message to logger."""
        super().__init__(*args, **kwargs)
        if args:
            logger.error(str(args[0]), stacklevel=2)


class gapsConfigError(gapsError):
    """gaps ConfigError."""


class gapsExecutionError(gapsError):
    """gaps ExecutionError."""


class gapsIndexError(gapsError, IndexError):
    """gaps IndexError."""


class gapsIOError(gapsError, IOError):
    """gaps IOError."""


class gapsKeyError(gapsError, KeyError):
    """gaps KeyError."""


class gapsRuntimeError(gapsError, RuntimeError):
    """gaps RuntimeError."""


class gapsTypeError(gapsError, TypeError):
    """gaps TypeError."""


class gapsValueError(gapsError, ValueError):
    """gaps ValueError."""


class gapsHPCError(gapsError):
    """gaps HPCError."""
